"""
Running Distributed Algorithms Across Many Celery Workers

Use this module to handle algorithm backtesting (tuning)
or for live trading.

Under the hood, this is a Celery task handler
that processes jobs from a broker's
messaging queue. This allows the Analysis Engine to process
many algorithmic workloads concurrently using Celery's horizontally-
scalable worker pool architecture.

**Publish a SPY 60-day Backtest to the Distributed Analysis Engine**

::

    ./tools/run-algo-with-publishing.sh SPY 60 -w

or manually with:

::

    use_date=$(date +"%Y-%m-%d")
    ticker_dataset="${ticker}-${use_date}.json"
    history_loc="s3://algohistory/${ticker_dataset}"
    report_loc="s3://algoreport/${ticker_dataset}"
    extract_loc="s3://algoready/${ticker_dataset}"
    backtest_loc="file:/tmp/${ticker_dataset}"

    sa -t ${ticker} \
        -p ${history_loc} \
        -o ${report_loc} \
        -e ${extract_loc} \
        -b ${backtest_loc} \
        -s ${backtest_start_date} \
        -n ${use_date} \
        -w

"""

import inspect
import types
import importlib.machinery
import datetime
import celery.task as celery_task
import analysis_engine.consts as ae_consts
import analysis_engine.build_result as build_result
import analysis_engine.work_tasks.custom_task as custom_task
import analysis_engine.run_algo as run_algo
import analysis_engine.algo as ae_algo  # base algo
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


@celery_task.task(
    bind=True,
    base=custom_task.CustomTask,
    queue='run_distributed_algorithm')
def run_distributed_algorithm(
        self,
        algo_req):
    """run_distributed_algorithm

    Process an Algorithm using a Celery task that is
    processed by a Celery worker

    :param algo_req: dictionary for key/values for
        running an algorithm using Celery workers
    """

    label = algo_req.get(
        'name',
        'ae-algo')
    verbose = algo_req.get(
        'verbose_task',
        False)
    debug = algo_req.get(
        'debug',
        False)

    # please be careful logging prod passwords:
    if debug:
        log.info(f'task - {label} - start algo_req={algo_req}')
    elif verbose:
        log.info(f'task - {label} - start ')
    # end of start log

    rec = {}
    res = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    created_algo_object = None
    custom_algo_module = None
    new_algo_object = None
    use_custom_algo = False
    found_algo_module = True  # assume the BaseAlgo
    should_publish_extract_dataset = False
    should_publish_history_dataset = False
    should_publish_report_dataset = False

    ticker = algo_req.get(
        'ticker',
        'SPY')
    num_days_back = algo_req.get(
        'num_days_back',
        75)
    name = algo_req.get(
        'name',
        'ae-algo')
    algo_module_path = algo_req.get(
        'mod_path',
        None)
    module_name = algo_req.get(
        'module_name',
        'BaseAlgo')
    custom_algo_module = algo_req.get(
        'custom_algo_module',
        None)
    new_algo_object = algo_req.get(
        'new_algo_object',
        None)
    use_custom_algo = algo_req.get(
        'use_custom_algo',
        False)
    should_publish_extract_dataset = algo_req.get(
        'should_publish_extract_dataset',
        False)
    should_publish_history_dataset = algo_req.get(
        'should_publish_history_dataset',
        False)
    should_publish_report_dataset = algo_req.get(
        'should_publish_report_dataset',
        False)
    start_date = algo_req.get(
        'start_date',
        None)
    end_date = algo_req.get(
        'end_date',
        None)
    raise_on_err = algo_req.get(
        'raise_on_err',
        True)
    report_config = algo_req.get(
        'report_config',
        None)
    history_config = algo_req.get(
        'history_config',
        None)
    extract_config = algo_req.get(
        'extract_config',
        None)

    err = None
    if algo_module_path:
        found_algo_module = False
        module_name = algo_module_path.split('/')[-1]
        loader = importlib.machinery.SourceFileLoader(
            module_name,
            algo_module_path)
        custom_algo_module = types.ModuleType(
            loader.name)
        loader.exec_module(
            custom_algo_module)
        use_custom_algo = True

        for member in inspect.getmembers(custom_algo_module):
            if module_name in str(member):
                found_algo_module = True
                break
        # for all members in this custom module file
    # if loading a custom algorithm module from a file on disk

    if not found_algo_module:
        err = (
            f'{label} - unable to find custom algorithm '
            f'module={custom_algo_module} module_path={algo_module_path}')
        if algo_module_path:
            err = (
                f'{label} - analysis_engine.'
                'work_tasks.run_distributed_algorithm was unable '
                f'to find custom algorithm module={custom_algo_module} with '
                f'provided path to \n file: {algo_module_path} \n'
                '\n'
                'Please confirm '
                'that the class inherits from the BaseAlgo class like:\n'
                '\n'
                'import analysis_engine.algo\n'
                'class MyAlgo(analysis_engine.algo.BaseAlgo):\n '
                '\n'
                'If it is then please file an issue on github:\n '
                'https://github.com/AlgoTraders/stock-analysis-engine/'
                'issues/new \n\nFor now this error results in a shutdown'
                '\n')
        # if algo_module_path set

        log.error(err)
        res = build_result.build_result(
            status=ae_consts.ERR,
            err=err,
            rec=None)
        task_result = {
            'status': res['status'],
            'err': res['err'],
            'algo_req': algo_req,
            'rec': rec
        }
        return task_result
    # if not found_algo_module

    use_start_date = start_date
    use_end_date = end_date
    if not use_end_date:
        end_date = datetime.datetime.utcnow()
        use_end_date = end_date.strftime(
            ae_consts.COMMON_TICK_DATE_FORMAT)
    if not use_start_date:
        start_date = end_date - datetime.timedelta(days=num_days_back)
        use_start_date = start_date.strftime(
            ae_consts.COMMON_TICK_DATE_FORMAT)
    dataset_publish_extract = algo_req.get(
        'dataset_publish_extract',
        False)
    dataset_publish_history = algo_req.get(
        'dataset_publish_history',
        False)
    dataset_publish_report = algo_req.get(
        'dataset_publish_report',
        False)
    try:
        if use_custom_algo:
            if verbose:
                log.info(
                    f'inspecting {custom_algo_module} for class {module_name}')
            use_class_member_object = None
            for member in inspect.getmembers(custom_algo_module):
                if module_name in str(member):
                    if verbose:
                        log.info(f'start {name} with {member[1]}')
                    use_class_member_object = member
                    break
            # end of looking over the class definition but did not find it

            if use_class_member_object:
                if algo_req.get('backtest', False):
                    new_algo_object = member[1](
                        ticker=algo_req['ticker'],
                        config_dict=algo_req)
                else:
                    new_algo_object = member[1](
                        **algo_req)
            else:
                err = (
                    f'{label} - did not find a derived '
                    'analysis_engine.algo.BaseAlgo '
                    f'class in the module file={algo_module_path} '
                    f'for ticker={ticker} algo_name={name}')
                log.error(err)
                res = build_result.build_result(
                    status=ae_consts.ERR,
                    err=err,
                    rec=None)
                task_result = {
                    'status': res['status'],
                    'err': res['err'],
                    'algo_req': algo_req,
                    'rec': rec
                }
                return task_result
            # end of finding a valid algorithm object
        else:
            new_algo_object = ae_algo.BaseAlgo(
                **algo_req)
        # if using a custom module path or the BaseAlgo

        if new_algo_object:
            # heads up - logging this might have passwords in the algo_req
            # log.debug(f'{name} algorithm request: {algo_req}')
            if verbose:
                log.info(
                    f'{name} - run START ticker={ticker} '
                    f'from {use_start_date} to {use_end_date}')
            if algo_req.get('backtest', False):
                algo_res = run_algo.run_algo(
                    algo=new_algo_object,
                    config_dict=algo_req)
                created_algo_object = new_algo_object
            else:
                algo_res = run_algo.run_algo(
                    algo=new_algo_object,
                    **algo_req)
                created_algo_object = new_algo_object

            if verbose:
                log.info(
                    f'{name} - run DONE ticker={ticker} '
                    f'from {use_start_date} to {use_end_date}')
            if debug:
                if custom_algo_module:
                    log.info(
                        f'{name} - done run_algo '
                        f'custom_algo_module={custom_algo_module} '
                        f'module_name={module_name} ticker={ticker} '
                        f'from {use_start_date} to {use_end_date}')
                else:
                    log.info(
                        f'{name} - done run_algo BaseAlgo ticker={ticker} '
                        f'from {use_start_date} to {use_end_date}')
        else:
            err = (
                f'{label} - missing a derived analysis_engine.algo.BaseAlgo '
                f'class in the module file={algo_module_path} for '
                f'ticker={ticker} algo_name={name}')
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=None)
            task_result = {
                'status': res['status'],
                'err': res['err'],
                'algo_req': algo_req,
                'rec': rec
            }
            return task_result
        # end of finding a valid algorithm object

        if not created_algo_object:
            err = (
                f'{label} - failed creating algorithm object - '
                f'ticker={ticker} '
                f'status={ae_consts.get_status(status=algo_res["status"])} '
                f'error={algo_res["err"]} algo name={name} '
                f'custom_algo_module={custom_algo_module} '
                f'module_name={module_name} '
                f'from {use_start_date} to {use_end_date}')
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=None)
            task_result = {
                'status': res['status'],
                'err': res['err'],
                'algo_req': algo_req,
                'rec': rec
            }
            return task_result
        # end of stop early

        if should_publish_extract_dataset or dataset_publish_extract:
            s3_log = ''
            redis_log = ''
            file_log = ''
            use_log = 'publish'

            if (extract_config['redis_address'] and
                    extract_config['redis_db'] and
                    extract_config['redis_key']):
                redis_log = (
                    f'redis://{extract_config["redis_address"]}'
                    f'@{extract_config["redis_db"]}'
                    f'/{extract_config["redis_key"]}')
                use_log += f' {redis_log}'
            else:
                extract_config['redis_enabled'] = False
            if (extract_config['s3_address'] and
                    extract_config['s3_bucket'] and
                    extract_config['s3_key']):
                s3_log = (
                    f's3://{extract_config["s3_address"]}'
                    f'/{extract_config["s3_bucket"]}'
                    f'/{extract_config["s3_key"]}')
                use_log += f' {s3_log}'
            else:
                extract_config['s3_enabled'] = False
            if extract_config['output_file']:
                file_log = f'file:{extract_config["output_file"]}'
                use_log += f' {file_log}'

            if verbose:
                log.info(
                    f'{name} - publish - start ticker={ticker} '
                    f'algorithm-ready {use_log}')

            publish_status = created_algo_object.publish_input_dataset(
                **extract_config)
            if publish_status != ae_consts.SUCCESS:
                msg = (
                    'failed to publish algorithm-ready datasets with '
                    f'status {ae_consts.get_status(status=publish_status)} '
                    f'attempted to {use_log}')
                log.error(msg)
                res = build_result.build_result(
                    status=ae_consts.ERR,
                    err=err,
                    rec=None)
                task_result = {
                    'status': res['status'],
                    'err': res['err'],
                    'algo_req': algo_req,
                    'rec': rec
                }
                return task_result
            # end of stop early

            if verbose:
                log.info(
                    f'{name} - publish - done ticker={ticker} '
                    f'algorithm-ready {use_log}')
        # if publish the algorithm-ready dataset

        if should_publish_history_dataset or dataset_publish_history:
            s3_log = ''
            redis_log = ''
            file_log = ''
            use_log = 'publish'

            if (history_config['redis_address'] and
                    history_config['redis_db'] and
                    history_config['redis_key']):
                redis_log = (
                    f'redis://{history_config["redis_address"]}'
                    f'@{history_config["redis_db"]}'
                    f'/{history_config["redis_key"]}')
                use_log += f' {redis_log}'
            if (history_config['s3_address'] and
                    history_config['s3_bucket'] and
                    history_config['s3_key']):
                s3_log = (
                    f's3://{history_config["s3_address"]}'
                    f'/{history_config["s3_bucket"]}'
                    f'/{history_config["s3_key"]}')
                use_log += f' {s3_log}'
            if history_config['output_file']:
                file_log = f'file:{history_config["output_file"]}'
                use_log += f' {file_log}'

            if verbose:
                log.info(
                    f'{name} - publish - start ticker={ticker} trading '
                    f'history {use_log}')

            publish_status = \
                created_algo_object.publish_trade_history_dataset(
                    **history_config)
            if publish_status != ae_consts.SUCCESS:
                msg = (
                    'failed to publish trading history datasets with '
                    f'status {ae_consts.get_status(status=publish_status)} '
                    f'attempted to {use_log}')
                log.error(msg)
                res = build_result.build_result(
                    status=ae_consts.ERR,
                    err=err,
                    rec=None)
                task_result = {
                    'status': res['status'],
                    'err': res['err'],
                    'algo_req': algo_req,
                    'rec': rec
                }
                return task_result
            # end of stop early

            if verbose:
                log.info(
                    f'{name} - publish - done ticker={ticker} trading '
                    f'history {use_log}')
        # if publish an trading history dataset

        if should_publish_report_dataset or dataset_publish_report:
            s3_log = ''
            redis_log = ''
            file_log = ''
            use_log = 'publish'

            if (report_config['redis_address'] and
                    report_config['redis_db'] and
                    report_config['redis_key']):
                redis_log = (
                    f'redis://{report_config["redis_address"]}'
                    f'@{report_config["redis_db"]}'
                    f'/{report_config["redis_key"]}')
                use_log += f' {redis_log}'
            if (report_config['s3_address'] and
                    report_config['s3_bucket'] and
                    report_config['s3_key']):
                s3_log = (
                    f's3://{report_config["s3_address"]}'
                    f'/{report_config["s3_bucket"]}'
                    f'/{report_config["s3_key"]}')
                use_log += f' {s3_log}'
            if report_config['output_file']:
                file_log = f' file:{report_config["output_file"]}'
                use_log += f' {file_log}'

            if verbose:
                log.info(
                    f'{name} - publishing ticker={ticker} trading performance '
                    f'report {use_log}')

            publish_status = created_algo_object.publish_report_dataset(
                **report_config)
            if publish_status != ae_consts.SUCCESS:
                msg = (
                    'failed to publish trading performance '
                    'report datasets with '
                    f'status {ae_consts.get_status(status=publish_status)} '
                    f'attempted to {use_log}')
                log.error(msg)
                res = build_result.build_result(
                    status=ae_consts.ERR,
                    err=err,
                    rec=None)
                task_result = {
                    'status': res['status'],
                    'err': res['err'],
                    'algo_req': algo_req,
                    'rec': rec
                }
                return task_result
            # end of stop early

            if verbose:
                log.info(
                    f'{name} - publish - done ticker={ticker} trading '
                    f'performance report {use_log}')
        # if publish an trading performance report dataset

        if verbose:
            log.info(
                f'{name} - done publishing datasets for ticker={ticker} '
                f'from {use_start_date} to {use_end_date}')

        rec['history_config'] = history_config
        rec['report_config'] = report_config

        res = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        res = build_result.build_result(
            status=ae_consts.ERR,
            err=(
                'failed - run_distributed_algorithm '
                f'dict={algo_req} with ex={e}'),
            rec=rec)
        if raise_on_err:
            raise e
        else:
            log.error(f'{label} - {res["err"]}')
    # end of try/ex

    if verbose:
        log.info(
            'task - run_distributed_algorithm done - '
            f'{label} - status={ae_consts.get_status(res["status"])}')

    task_result = {
        'status': res['status'],
        'err': res['err'],
        'algo_req': algo_req,
        'rec': rec
    }
    return task_result
# end of run_distributed_algorithm



"""
Helper for loading datasets from redis

**Supported environment variables**

::

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json
"""

import analysis_engine.consts as ae_consts
import analysis_engine.get_data_from_redis_key as redis_utils
import analysis_engine.prepare_dict_for_algo as prepare_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def load_algo_dataset_from_redis(
        redis_key,
        redis_address,
        redis_db,
        redis_password,
        redis_expire=None,
        redis_serializer='json',
        serialize_datasets=ae_consts.DEFAULT_SERIALIZED_DATASETS,
        compress=False,
        encoding='utf-8'):
    """load_algo_dataset_from_redis

    Load an algorithm-ready dataset for algorithm backtesting
    from a redis key

    :param serialize_datasets: optional - list of dataset names to
        deserialize in the dataset
    :param compress: optional - boolean flag for decompressing
        the contents of the ``path_to_file`` if necessary
        (default is ``False`` and algorithms
        use ``zlib`` for compression)
    :param encoding: optional - string for data encoding
    """
    log.debug('start')
    data_from_file = None

    redis_host = redis_address.split(':')[0]
    redis_port = int(redis_address.split(':')[0])

    redis_res = redis_utils.get_data_from_redis_key(
        key=redis_key,
        host=redis_host,
        port=redis_port,
        db=redis_db,
        password=redis_password,
        expire=redis_expire,
        serializer=redis_serializer,
        encoding=encoding)

    if redis_res['status'] != ae_consts.SUCCESS:
        log.error(
            'failed getting data from '
            f'redis={redis_address}:{redis_db}/{redis_key}')
        return None

    data_from_file = redis_res['rec']['data']
    if not data_from_file:
        log.error(
            f'missing data from redis={redis_address}:{redis_db}/{redis_key}')
        return None

    return prepare_utils.prepare_dict_for_algo(
        data=data_from_file,
        compress=compress,
        convert_to_dict=True,
        encoding=encoding)
# end of load_algo_dataset_from_redis



"""
Helper for loading datasets from s3
"""

import boto3
import analysis_engine.consts as ae_consts
import analysis_engine.prepare_dict_for_algo as prepare_utils
import analysis_engine.s3_read_contents_from_key as s3_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def load_algo_dataset_from_s3(
        s3_key,
        s3_address,
        s3_bucket,
        s3_access_key,
        s3_secret_key,
        s3_region_name,
        s3_secure,
        serialize_datasets=ae_consts.DEFAULT_SERIALIZED_DATASETS,
        compress=False,
        encoding='utf-8'):
    """load_algo_dataset_from_s3

    Load an algorithm-ready dataset for algorithm backtesting
    from a local file

    :param serialize_datasets: optional - list of dataset names to
        deserialize in the dataset
    :param compress: optional - boolean flag for decompressing
        the contents of the ``path_to_file`` if necessary
        (default is ``False`` and algorithms
        use ``zlib`` for compression)
    :param encoding: optional - string for data encoding

    **Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``True``)
    :param s3_key: string - key to save the data in redis
        (default is ``None``)
    :param s3_address: Minio S3 connection string format: ``host:port``
        (default is ``localhost:9000``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``dev``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``trexaccesskey``)
    :param s3_secret_key: S3 Secret key
        (default is ``trex123321``)
    :param s3_region_name: S3 region name
        (default is ``us-east-1``)
    :param s3_secure: Transmit using tls encryption
        (default is ``False``)
    """
    log.info(
        f'start s3={s3_address}:{s3_bucket}/{s3_key}')

    data_from_file = None

    endpoint_url = f'http://{s3_address}'
    if s3_secure:
        endpoint_url = f'https://{s3_address}'

    s3 = boto3.resource(
        's3',
        endpoint_url=endpoint_url,
        aws_access_key_id=s3_access_key,
        aws_secret_access_key=s3_secret_key,
        region_name=s3_region_name,
        config=boto3.session.Config(signature_version='s3v4'))

    # compressed files will not work with json.dumps
    try:
        data_from_file = s3_utils.s3_read_contents_from_key(
            s3=s3,
            s3_bucket_name=s3_bucket,
            s3_key=s3_key,
            encoding=encoding,
            convert_as_json=not compress,
            compress=compress)
    except Exception as e:
        if (
                'An error occurred (NoSuchBucket) '
                'when calling the GetObject operation') in str(e):
            msg = (
                f'missing s3_bucket={s3_address} in s3_address={s3_bucket}')
            log.error(msg)
            raise Exception(msg)
        else:
            raise Exception(e)

    if not data_from_file:
        log.error(
            'missing data from s3={s3_address}:{s3_bucket}/{s3_key}')
        return None

    return prepare_utils.prepare_dict_for_algo(
        data=data_from_file,
        compress=False,
        convert_to_dict=True,
        encoding=encoding)
# end of load_algo_dataset_from_s3



#!/usr/bin/env python

"""
Algorithm Runner API Example Script

**Run Full Backtest**

::

    backtest_with_runner.py -t TICKER -b S3_BUCKET -k S3_KEY -c ALGO_CONFIG

**Run Algorithm with Latest Pricing Data**

::

    backtest_with_runner.py -l -t TICKER -b S3_BUCKET -k S3_KEY -c ALGO_CONFIG

Debug by adding ``-d`` as an argument
"""

import sys
import argparse
import analysis_engine.utils as ae_utils
import analysis_engine.algo_runner as algo_runner
import analysis_engine.plot_trading_history as plot
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(
    name='algo-runner')


def backtest_with_runner():
    """backtest_with_runner

    build and publish a trading history from an algorithm config.

    ::

        backtest_with_runner.py -t TICKER -c ALGO_CONFIG -s START_DATE
        -k S3_KEY -b S3_BUCKET -l
    """

    parser = argparse.ArgumentParser(
        description=(
            'backtest an algorithm and publish '
            'the trading history'))
    parser.add_argument(
        '-t',
        help=('ticker symbol'),
        required=False,
        dest='ticker')
    parser.add_argument(
        '-k',
        help=('s3_key'),
        required=False,
        dest='s3_key')
    parser.add_argument(
        '-b',
        help=('s3_bucket'),
        required=False,
        dest='s3_bucket')
    parser.add_argument(
        '-s',
        help=('start date format YYYY-MM-DD'),
        required=False,
        dest='start_date')
    parser.add_argument(
        '-c',
        help=('algo config file'),
        required=False,
        dest='algo_config')
    parser.add_argument(
        '-l',
        help=(
            'run a backtest with the latest '
            'pricing data'),
        required=False,
        dest='latest',
        action='store_true')
    parser.add_argument(
        '-d',
        help='debug',
        required=False,
        dest='debug',
        action='store_true')
    args = parser.parse_args()

    ticker = 'SPY'
    s3_bucket = (
        f'algohistory')
    s3_key = (
        f'trade_history_{ticker}')
    start_date = (
        f'2019-01-01')
    algo_config = (
        f'/opt/sa/cfg/default_algo.json')
    latest = False
    show_plot = True
    debug = False

    if args.ticker:
        ticker = args.ticker.upper()
    if args.s3_key:
        s3_key = args.s3_key
    if args.s3_bucket:
        s3_bucket = args.s3_bucket
    if args.start_date:
        start_date = args.start_date
    if args.algo_config:
        algo_config = args.algo_config
    if args.latest:
        latest = True
        start_date = ae_utils.get_last_close_str()
    if args.debug:
        debug = True

    history_loc = (
        f's3://{s3_bucket}/{s3_key}')

    log.info(
        f'building {ticker} trade history '
        f'start_date={start_date} '
        f'config={algo_config} '
        f'history_loc={history_loc}')

    runner = algo_runner.AlgoRunner(
        ticker=ticker,
        start_date=start_date,
        history_loc=history_loc,
        algo_config=algo_config,
        verbose_algo=debug,
        verbose_processor=False,
        verbose_indicators=False)

    trading_history_df = None
    if latest:
        trading_history_df = runner.latest()
        log.info(
            f'{ticker} latest:')
        print(trading_history_df[['minute', 'close']].tail(5))
        log.info(
            f'Other available columns to plot:')
        print(trading_history_df.columns.values)
        if show_plot:
            plot.plot_trading_history(
                title=(
                    f'{ticker} at '
                    f'${trading_history_df["close"].iloc[-1]} '
                    f'at: '
                    f'{trading_history_df["minute"].iloc[-1]}'),
                df=trading_history_df,
                red='high',
                blue='close')
    else:
        runner.start()

    sys.exit(0)
# end of backtest_with_runner


if __name__ == '__main__':
    backtest_with_runner()



#!/usr/bin/env python

"""
A tool for showing how to build an algorithm and
run a backtest with an algorithm config dictionary

.. code-block:: python

    import analysis_engine.consts as ae_consts
    import analysis_engine.algo as base_algo
    import analysis_engine.run_algo as run_algo

    ticker = 'SPY'

    willr_close_path = (
        'analysis_engine/mocks/example_indicator_williamsr.py')
    willr_open_path = (
        'analysis_engine/mocks/example_indicator_williamsr_open.py')
    algo_config_dict = {
        'name': 'min-runner',
        'timeseries': timeseries,
        'trade_horizon': 5,
        'num_owned': 10,
        'buy_shares': 10,
        'balance': 10000.0,
        'commission': 6.0,
        'ticker': ticker,
        'algo_module_path': None,
        'algo_version': 1,
        'verbose': False,               # log in the algorithm
        'verbose_processor': False,     # log in the indicator processor
        'verbose_indicators': False,    # log all indicators
        'verbose_trading': True,        # log in the algo trading methods
        'positions': {
            ticker: {
                'shares': 10,
                'buys': [],
                'sells': []
            }
        },
        'buy_rules': {
            'confidence': 75,
            'min_indicators': 3
        },
        'sell_rules': {
            'confidence': 75,
            'min_indicators': 3
        },
        'indicators': [
            {
                'name': 'willr_-70_-30',
                'module_path': willr_close_path,
                'category': 'technical',
                'type': 'momentum',
                'uses_data': 'minute',
                'high': 0,
                'low': 0,
                'close': 0,
                'open': 0,
                'willr_value': 0,
                'num_points': 80,
                'buy_below': -70,
                'sell_above': -30,
                'is_buy': False,
                'is_sell': False,
                'verbose': False  # log in just this indicator
            },
            {
                'name': 'willr_-80_-20',
                'module_path': willr_close_path,
                'category': 'technical',
                'type': 'momentum',
                'uses_data': 'minute',
                'high': 0,
                'low': 0,
                'close': 0,
                'open': 0,
                'willr_value': 0,
                'num_points': 30,
                'buy_below': -80,
                'sell_above': -20,
                'is_buy': False,
                'is_sell': False
            },
            {
                'name': 'willr_-90_-10',
                'module_path': willr_close_path,
                'category': 'technical',
                'type': 'momentum',
                'uses_data': 'minute',
                'high': 0,
                'low': 0,
                'close': 0,
                'open': 0,
                'willr_value': 0,
                'num_points': 60,
                'buy_below': -90,
                'sell_above': -10,
                'is_buy': False,
                'is_sell': False
            },
            {
                'name': 'willr_open_-80_-20',
                'module_path': willr_open_path,
                'category': 'technical',
                'type': 'momentum',
                'uses_data': 'minute',
                'high': 0,
                'low': 0,
                'close': 0,
                'open': 0,
                'willr_open_value': 0,
                'num_points': 80,
                'buy_below': -80,
                'sell_above': -20,
                'is_buy': False,
                'is_sell': False
            }
        ],
        'slack': {
            'webhook': None
        }
    }

    class ExampleCustomAlgo(base_algo.BaseAlgo):
        def process(self, algo_id, ticker, dataset):
            if self.verbose:
                print(
                    f'process start - {self.name} '
                    f'date={self.backtest_date} minute={self.latest_min} '
                    f'close={self.latest_close} high={self.latest_high} '
                    f'low={self.latest_low} open={self.latest_open} '
                    f'volume={self.latest_volume}')
        # end of process
    # end of ExampleCustomAlgo


    algo_obj = ExampleCustomAlgo(
        ticker=algo_config_dict['ticker'],
        config_dict=algo_config_dict)

    algo_res = run_algo.run_algo(
        ticker=algo_config_dict['ticker'],
        algo=algo_obj,
        raise_on_err=True)

    if algo_res['status'] != ae_consts.SUCCESS:
        print(
            'failed running algo backtest '
            f'{algo_obj.get_name()} hit status: '
            f'{ae_consts.get_status(status=algo_res['status'])} '
            f'error: {algo_res["err"]}')
    else:
        print(
            f'backtest: {algo_obj.get_name()} '
            f'{ae_consts.get_status(status=algo_res["status"])} - '
            'plotting history')
    # if not successful

"""

import os
import sys
import datetime
import argparse
import analysis_engine.consts as ae_consts
import analysis_engine.algo as base_algo
import analysis_engine.run_algo as run_algo
import analysis_engine.plot_trading_history as plot_trading_history
import analysis_engine.build_publish_request as build_publish_request
import spylunking.log.setup_logging as log_utils


log = log_utils.build_colorized_logger(
    name='bt',
    log_config_path=ae_consts.LOG_CONFIG_PATH)


def build_example_algo_config(
        ticker,
        timeseries='minute'):
    """build_example_algo_config

    helper for building an algorithm config dictionary

    :returns: algorithm config dictionary
    """
    willr_close_path = (
        'analysis_engine/mocks/example_indicator_williamsr.py')
    willr_open_path = (
        'analysis_engine/mocks/example_indicator_williamsr_open.py')
    algo_config_dict = {
        'name': 'backtest',
        'timeseries': timeseries,
        'trade_horizon': 5,
        'num_owned': 10,
        'buy_shares': 10,
        'balance': 10000.0,
        'commission': 6.0,
        'ticker': ticker,
        'algo_module_path': None,
        'algo_version': 1,
        'verbose': False,  # log in the algorithm
        'verbose_processor': False,  # log in the indicator processor
        'verbose_indicators': False,  # log all indicators
        'verbose_trading': False,  # log in the algo trading methods
        'inspect_datasets': False,  # log dataset metrics - slow
        'positions': {
            ticker: {
                'shares': 10,
                'buys': [],
                'sells': []
            }
        },
        'buy_rules': {
            'confidence': 75,
            'min_indicators': 3
        },
        'sell_rules': {
            'confidence': 75,
            'min_indicators': 3
        },
        'indicators': [
            {
                'name': 'willr_-70_-30',
                'module_path': willr_close_path,
                'category': 'technical',
                'type': 'momentum',
                'uses_data': 'minute',
                'high': 0,
                'low': 0,
                'close': 0,
                'open': 0,
                'willr_value': 0,
                'num_points': 80,
                'buy_below': -70,
                'sell_above': -30,
                'is_buy': False,
                'is_sell': False,
                'verbose': False  # log in just this indicator
            },
            {
                'name': 'willr_-80_-20',
                'module_path': willr_close_path,
                'category': 'technical',
                'type': 'momentum',
                'uses_data': 'minute',
                'high': 0,
                'low': 0,
                'close': 0,
                'open': 0,
                'willr_value': 0,
                'num_points': 30,
                'buy_below': -80,
                'sell_above': -20,
                'is_buy': False,
                'is_sell': False
            },
            {
                'name': 'willr_-90_-10',
                'module_path': willr_close_path,
                'category': 'technical',
                'type': 'momentum',
                'uses_data': 'minute',
                'high': 0,
                'low': 0,
                'close': 0,
                'open': 0,
                'willr_value': 0,
                'num_points': 60,
                'buy_below': -90,
                'sell_above': -10,
                'is_buy': False,
                'is_sell': False
            },
            {
                'name': 'willr_open_-80_-20',
                'module_path': willr_open_path,
                'category': 'technical',
                'type': 'momentum',
                'uses_data': 'minute',
                'high': 0,
                'low': 0,
                'close': 0,
                'open': 0,
                'willr_open_value': 0,
                'num_points': 80,
                'buy_below': -80,
                'sell_above': -20,
                'is_buy': False,
                'is_sell': False
            }
        ],
        'slack': {
            'webhook': None
        }
    }

    return algo_config_dict
# end of build_example_algo_config


class ExampleCustomAlgo(base_algo.BaseAlgo):
    """ExampleCustomAlgo"""

    def process(self, algo_id, ticker, dataset):
        """process

        Run a custom algorithm after all the indicators
        from the ``algo_config_dict`` have been processed and all
        the number crunching is done. This allows the algorithm
        class to focus on the high-level trade execution problems
        like bid-ask spreads and opening the buy/sell trade orders.

        **How does it work?**

        The engine provides a data stream from the latest
        pricing updates stored in redis. Once new data is
        stored in redis, algorithms will be able to use
        each ``dataset`` as a chance to evaluate buy and
        sell decisions. These are your own custom logic
        for trading based off what the indicators find
        and any non-indicator data provided from within
        the ``dataset`` dictionary.

        **Dataset Dictionary Structure**

        Here is what the ``dataset`` variable
        looks like when your algorithm's ``process``
        method is called (assuming you have redis running
        with actual pricing data too):

        .. code-block:: python

            dataset = {
                'id': dataset_id,
                'date': date,
                'data': {
                    'daily': pd.DataFrame([]),
                    'minute': pd.DataFrame([]),
                    'quote': pd.DataFrame([]),
                    'stats': pd.DataFrame([]),
                    'peers': pd.DataFrame([]),
                    'news1': pd.DataFrame([]),
                    'financials': pd.DataFrame([]),
                    'earnings': pd.DataFrame([]),
                    'dividends': pd.DataFrame([]),
                    'calls': pd.DataFrame([]),
                    'puts': pd.DataFrame([]),
                    'pricing': pd.DataFrame([]),
                    'news': pd.DataFrame([])
                }
            }

        .. tip:: you can also inspect these datasets by setting
            the algorithm's config dictionary key
            ``"inspect_datasets": True``

        :param algo_id: string - algo identifier label for debugging datasets
            during specific dates
        :param ticker: string - ticker
        :param dataset: a dictionary of identifiers (for debugging) and
            multiple pandas ``DataFrame`` objects.
        """
        if self.verbose:
            log.info(
                f'process start - {self.name} balance={self.balance} '
                f'date={self.backtest_date} minute={self.latest_min} '
                f'close={self.latest_close} high={self.latest_high} '
                f'low={self.latest_low} open={self.latest_open} '
                f'volume={self.latest_volume}')
    # end of process

# end of ExampleCustomAlgo


def run_backtest_and_plot_history(
        config_dict):
    """run_backtest_and_plot_history

    Run a derived algorithm with an algorithm config dictionary

    :param config_dict: algorithm config dictionary
    """

    log.debug('start - sa')

    parser = argparse.ArgumentParser(
        description=(
            'stock analysis tool'))
    parser.add_argument(
        '-t',
        help=(
            'ticker'),
        required=True,
        dest='ticker')
    parser.add_argument(
        '-e',
        help=(
            'file path to extract an '
            'algorithm-ready datasets from redis'),
        required=False,
        dest='algo_extract_loc')
    parser.add_argument(
        '-l',
        help=(
            'show dataset in this file'),
        required=False,
        dest='show_from_file')
    parser.add_argument(
        '-H',
        help=(
            'show trading history dataset in this file'),
        required=False,
        dest='show_history_from_file')
    parser.add_argument(
        '-E',
        help=(
            'show trading performance report dataset in this file'),
        required=False,
        dest='show_report_from_file')
    parser.add_argument(
        '-L',
        help=(
            'restore an algorithm-ready dataset file back into redis'),
        required=False,
        dest='restore_algo_file')
    parser.add_argument(
        '-f',
        help=(
            'save the trading history dataframe '
            'to this file'),
        required=False,
        dest='history_json_file')
    parser.add_argument(
        '-J',
        help=(
            'plot action - after preparing you can use: '
            '-J show to open the image (good for debugging)'),
        required=False,
        dest='plot_action')
    parser.add_argument(
        '-b',
        help=(
            'run a backtest using the dataset in '
            'a file path/s3 key/redis key formats: '
            'file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
            's3://algoready/SPY-latest.json or '
            'redis:SPY-latest'),
        required=False,
        dest='backtest_loc')
    parser.add_argument(
        '-B',
        help=(
            'optional - broker url for Celery'),
        required=False,
        dest='broker_url')
    parser.add_argument(
        '-C',
        help=(
            'optional - broker url for Celery'),
        required=False,
        dest='backend_url')
    parser.add_argument(
        '-w',
        help=(
            'optional - flag for publishing an algorithm job '
            'using Celery to the ae workers'),
        required=False,
        dest='run_on_engine',
        action='store_true')
    parser.add_argument(
        '-k',
        help=(
            'optional - s3 access key'),
        required=False,
        dest='s3_access_key')
    parser.add_argument(
        '-K',
        help=(
            'optional - s3 secret key'),
        required=False,
        dest='s3_secret_key')
    parser.add_argument(
        '-a',
        help=(
            'optional - s3 address format: <host:port>'),
        required=False,
        dest='s3_address')
    parser.add_argument(
        '-Z',
        help=(
            'optional - s3 secure: default False'),
        required=False,
        dest='s3_secure')
    parser.add_argument(
        '-s',
        help=(
            'optional - start date: YYYY-MM-DD'),
        required=False,
        dest='start_date')
    parser.add_argument(
        '-n',
        help=(
            'optional - end date: YYYY-MM-DD'),
        required=False,
        dest='end_date')
    parser.add_argument(
        '-u',
        help=(
            'optional - s3 bucket name'),
        required=False,
        dest='s3_bucket_name')
    parser.add_argument(
        '-G',
        help=(
            'optional - s3 region name'),
        required=False,
        dest='s3_region_name')
    parser.add_argument(
        '-g',
        help=(
            'Path to a custom algorithm module file '
            'on disk. This module must have a single '
            'class that inherits from: '
            'https://github.com/AlgoTraders/stock-analysis-engine/'
            'blob/master/'
            'analysis_engine/algo.py Additionally you '
            'can find the Example-Minute-Algorithm here: '
            'https://github.com/AlgoTraders/stock-analysis-engine/'
            'blob/master/analysis_engine/mocks/'
            'example_algo_minute.py'),
        required=False,
        dest='run_algo_in_file')
    parser.add_argument(
        '-p',
        help=(
            'optional - s3 bucket/file for trading history'),
        required=False,
        dest='algo_history_loc')
    parser.add_argument(
        '-o',
        help=(
            'optional - s3 bucket/file for trading performance report'),
        required=False,
        dest='algo_report_loc')
    parser.add_argument(
        '-r',
        help=(
            'optional - redis_address format: <host:port>'),
        required=False,
        dest='redis_address')
    parser.add_argument(
        '-R',
        help=(
            'optional - redis and s3 key name'),
        required=False,
        dest='keyname')
    parser.add_argument(
        '-m',
        help=(
            'optional - redis database number (0 by default)'),
        required=False,
        dest='redis_db')
    parser.add_argument(
        '-x',
        help=(
            'optional - redis expiration in seconds'),
        required=False,
        dest='redis_expire')
    parser.add_argument(
        '-c',
        help=(
            'optional - algorithm config_file path for setting '
            'up internal algorithm trading strategies and '
            'indicators'),
        required=False,
        dest='config_file')
    parser.add_argument(
        '-v',
        help=(
            'set the Algorithm to verbose logging'),
        required=False,
        dest='verbose_algo',
        action='store_true')
    parser.add_argument(
        '-P',
        help=(
            'set the Algorithm\'s IndicatorProcessor to verbose logging'),
        required=False,
        dest='verbose_processor',
        action='store_true')
    parser.add_argument(
        '-I',
        help=(
            'set all Algorithm\'s Indicators to verbose logging '
            '(note indivdual indicators support a \'verbose\' key '
            'that can be set to True to debug just one '
            'indicator)'),
        required=False,
        dest='verbose_indicators',
        action='store_true')
    parser.add_argument(
        '-V',
        help=(
            'inspect the datasets an algorithm is processing - this'
            'will slow down processing to show debugging'),
        required=False,
        dest='inspect_datasets',
        action='store_true')
    parser.add_argument(
        '-j',
        help=(
            'run the algorithm on just this specific date in the datasets '
            '- specify the date in a format: YYYY-MM-DD like: 2018-11-29'),
        required=False,
        dest='run_this_date')
    parser.add_argument(
        '-d',
        help=(
            'debug'),
        required=False,
        dest='debug',
        action='store_true')
    args = parser.parse_args()

    ticker = ae_consts.TICKER
    use_balance = 10000.0
    use_commission = 6.0
    use_start_date = None
    use_end_date = None
    use_config_file = None
    debug = False
    verbose_algo = None
    verbose_processor = None
    verbose_indicators = None
    inspect_datasets = None
    history_json_file = None
    run_this_date = None

    s3_access_key = ae_consts.S3_ACCESS_KEY
    s3_secret_key = ae_consts.S3_SECRET_KEY
    s3_region_name = ae_consts.S3_REGION_NAME
    s3_address = ae_consts.S3_ADDRESS
    s3_secure = ae_consts.S3_SECURE
    redis_address = ae_consts.REDIS_ADDRESS
    redis_password = ae_consts.REDIS_PASSWORD
    redis_db = ae_consts.REDIS_DB
    redis_expire = ae_consts.REDIS_EXPIRE

    if args.s3_access_key:
        s3_access_key = args.s3_access_key
    if args.s3_secret_key:
        s3_secret_key = args.s3_secret_key
    if args.s3_region_name:
        s3_region_name = args.s3_region_name
    if args.s3_address:
        s3_address = args.s3_address
    if args.s3_secure:
        s3_secure = args.s3_secure
    if args.redis_address:
        redis_address = args.redis_address
    if args.redis_db:
        redis_db = args.redis_db
    if args.redis_expire:
        redis_expire = args.redis_expire
    if args.history_json_file:
        history_json_file = args.history_json_file
    if args.ticker:
        ticker = args.ticker.upper()
    if args.debug:
        debug = True
    if args.verbose_algo:
        verbose_algo = True
    if args.verbose_processor:
        verbose_processor = True
    if args.verbose_indicators:
        verbose_indicators = True
    if args.inspect_datasets:
        inspect_datasets = True
    if args.run_this_date:
        run_this_date = args.run_this_date

    if args.start_date:
        try:
            use_start_date = f'{str(args.start_date)} 00:00:00'
            datetime.datetime.strptime(
                args.start_date,
                ae_consts.COMMON_DATE_FORMAT)
        except Exception as e:
            msg = (
                'please use a start date formatted as: '
                f'{ae_consts.COMMON_DATE_FORMAT}\nerror was: {e}')
            log.error(msg)
            sys.exit(1)
        # end of testing for a valid date
    # end of args.start_date
    if args.end_date:
        try:
            use_end_date = f'{str(args.end_date)} 00:00:00'
            datetime.datetime.strptime(
                args.end_date,
                ae_consts.COMMON_DATE_FORMAT)
        except Exception as e:
            msg = (
                'please use an end date formatted as: '
                f'{ae_consts.COMMON_DATE_FORMAT}\nerror was: {e}')
            log.error(msg)
            sys.exit(1)
        # end of testing for a valid date
    # end of args.end_date
    if args.config_file:
        use_config_file = args.config_file
        if not os.path.exists(use_config_file):
            log.error(
                f'Failed: unable to find config file: -c {use_config_file}')
            sys.exit(1)

    if args.backtest_loc:
        backtest_loc = args.backtest_loc
        if ('file:/' not in backtest_loc and
                's3://' not in backtest_loc and
                'redis://' not in backtest_loc):
            log.error(
                'invalid -b <backtest dataset file> specified. '
                f'{backtest_loc} '
                'please use either: '
                '-b file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                '-b s3://algoready/SPY-latest.json or '
                '-b redis://SPY-latest')
            sys.exit(1)

        load_from_s3_bucket = None
        load_from_s3_key = None
        load_from_redis_key = None
        load_from_file = None

        if 's3://' in backtest_loc:
            load_from_s3_bucket = backtest_loc.split('/')[-2]
            load_from_s3_key = backtest_loc.split('/')[-1]
        elif 'redis://' in backtest_loc:
            load_from_redis_key = backtest_loc.split('/')[-1]
        elif 'file:/' in backtest_loc:
            load_from_file = backtest_loc.split(':')[-1]
        # end of parsing supported transport - loading an algo-ready

        load_config = build_publish_request.build_publish_request(
            ticker=ticker,
            output_file=load_from_file,
            s3_bucket=load_from_s3_bucket,
            s3_key=load_from_s3_key,
            redis_key=load_from_redis_key,
            redis_address=redis_address,
            redis_db=redis_db,
            redis_password=redis_password,
            redis_expire=redis_expire,
            s3_address=s3_address,
            s3_access_key=s3_access_key,
            s3_secret_key=s3_secret_key,
            s3_region_name=s3_region_name,
            s3_secure=s3_secure,
            verbose=debug,
            label=f'load-{backtest_loc}')
        if load_from_file:
            load_config['output_file'] = load_from_file
        if load_from_redis_key:
            load_config['redis_key'] = load_from_redis_key
            load_config['redis_enabled'] = True
        if load_from_s3_bucket and load_from_s3_key:
            load_config['s3_bucket'] = load_from_s3_bucket
            load_config['s3_key'] = load_from_s3_key
            load_config['s3_enabled'] = True

    if debug:
        log.info('starting algo')

    config_dict['ticker'] = ticker
    config_dict['balance'] = use_balance
    config_dict['commission'] = use_commission

    if verbose_algo:
        config_dict['verbose'] = verbose_algo
    if verbose_processor:
        config_dict['verbose_processor'] = verbose_processor
    if verbose_indicators:
        config_dict['verbose_indicators'] = verbose_indicators
    if inspect_datasets:
        config_dict['inspect_datasets'] = inspect_datasets
    if run_this_date:
        config_dict['run_this_date'] = run_this_date

    algo_obj = ExampleCustomAlgo(
        ticker=config_dict['ticker'],
        config_dict=config_dict)

    algo_res = run_algo.run_algo(
        ticker=ticker,
        algo=algo_obj,
        start_date=use_start_date,
        end_date=use_end_date,
        raise_on_err=True)

    if algo_res['status'] != ae_consts.SUCCESS:
        log.error(
            'failed running algo backtest '
            f'{algo_obj.get_name()} hit status: '
            f'{ae_consts.get_status(status=algo_res["status"])} '
            f'error: {algo_res["err"]}')
        return
    # if not successful

    log.info(
        f'backtest: {algo_obj.get_name()} '
        f'{ae_consts.get_status(status=algo_res["status"])}')

    trading_history_dict = algo_obj.get_history_dataset()
    history_df = trading_history_dict[ticker]
    if not hasattr(history_df, 'to_json'):
        return

    if history_json_file:
        log.info(f'saving history to: {history_json_file}')
        history_df.to_json(
            history_json_file,
            orient='records',
            date_format='iso')

    log.info('plotting history')

    use_xcol = 'date'
    use_as_date_format = '%d\n%b'
    xlabel = f'Dates vs {trading_history_dict["algo_name"]} values'
    ylabel = f'Algo {trading_history_dict["algo_name"]}\nvalues'
    df_filter = (history_df['close'] > 0.01)
    first_date = history_df[df_filter]['date'].iloc[0]
    end_date = history_df[df_filter]['date'].iloc[-1]
    if config_dict['timeseries'] == 'minute':
        use_xcol = 'minute'
        use_as_date_format = '%d %H:%M:%S\n%b'
        first_date = history_df[df_filter]['minute'].iloc[0]
        end_date = history_df[df_filter]['minute'].iloc[-1]
    title = (
        f'Trading History {ticker} for Algo '
        f'{trading_history_dict["algo_name"]}\n'
        f'Backtest dates from {first_date} to {end_date}')

    # set default hloc columns:
    blue = None
    green = None
    orange = None

    red = 'close'
    blue = 'balance'

    if debug:
        for i, r in history_df.iterrows():
            log.debug(f'{r["minute"]} - {r["close"]}')

    plot_trading_history.plot_trading_history(
        title=title,
        df=history_df,
        red=red,
        blue=blue,
        green=green,
        orange=orange,
        date_col=use_xcol,
        date_format=use_as_date_format,
        xlabel=xlabel,
        ylabel=ylabel,
        df_filter=df_filter,
        show_plot=True,
        dropna_for_all=True)

# end of run_backtest_and_plot_history


def start_backtest_with_plot_history():
    """start_backtest_with_plot_history

    setup.py helper for kicking off a backtest
    that will plot the trading history using
    seaborn and matplotlib showing
    the algorithm's balance vs the closing price
    of the asset
    """
    run_backtest_and_plot_history(
        config_dict=build_example_algo_config(
            ticker='SPY',
            timeseries='minute'))
# end of start_backtest_with_plot_history


if __name__ == '__main__':
    start_backtest_with_plot_history()



"""
A class for running backtests and the latest pricing data
with an automated publishing of the ``Trading History`` to S3
"""

import os
import datetime
import json
import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.build_dataset_node as build_dataset_node
import analysis_engine.load_history_dataset as load_history_utils
import analysis_engine.run_custom_algo as run_custom_algo
import analysis_engine.publish as publish
import analysis_engine.algo as base_algo
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


class AlgoRunner:
    """AlgoRunner

    Run an algorithm backtest or with the latest
    pricing data and publish the compressed
    trading history to s3 which can be used
    to train AI

    **Full Backtest**

    .. code-block:: python

        import analysis_engine.algo_runner as algo_runner
        runner = algo_runner.AlgoRunner('SPY')
        runner.start()

    **Run Algorithm with Latest Pricing Data**

    .. code-block:: python

        import analysis_engine.algo_runner as algo_runner
        import analysis_engine.plot_trading_history as plot
        ticker = 'SPY'
        runner = algo_runner.AlgoRunner(ticker)
        # run the algorithm with the latest 200 minutes:
        df = runner.latest()
        print(df[['minute', 'close']].tail(5))
        plot.plot_trading_history(
            title=(
                f'{ticker} - ${df["close"].iloc[-1]} '
                f'at: {df["minute"].iloc[-1]}'),
            df=df)

    """

    def __init__(
            self,
            ticker,
            algo_config=None,
            start_date=None,
            end_date=None,
            history_loc=None,
            predictions_loc=None,
            run_on_engine=False,
            verbose_algo=False,
            verbose_processor=False,
            verbose_indicators=False,
            **kwargs):
        """__init__

        constructor

        :param ticker: string ticker
        :param algo_config: optional - string path to file
            (default is ``./cfg/default_algo.json``)
        :param history_loc: optional - string trading history location
            (default is ``s3://algohistory/trade_history_{ticker}``)
        :param predictions_loc: optional - string predictions location
        :param start_date: optional - string start date
        :param end_date: optional - string end date
        :param run_on_engine: optional - bool flag for running local
            or distributed on an engine
            (default is ``False`` - local)
        :param verbose_algo: optional - bool flag for
            debugging the algo
            (default is ``False``)
        :param verbose_processor: optional - bool flag for
            debugging the algo's indicator processor
            (default is ``False``)
        :param verbose_indicators: optional - bool flag for
            debugging the algo's indicators
            (default is ``False``)
        :param kwargs: keyword args dictionary
        """
        self.ticker = None
        if ticker:
            self.ticker = str(ticker).upper()
        self.start_date = start_date
        self.end_date = end_date
        self.start_day = None
        self.end_day = None
        self.run_on_engine = run_on_engine
        self.algo_history_loc = (
            f's3://algohistory/trade_history_{self.ticker}')
        self.algo_predictions_loc = (
            f's3://predictions/{self.ticker}')
        if history_loc:
            self.algo_history_loc = history_loc
        if predictions_loc:
            self.algo_predictions_loc = predictions_loc

        self.pt_s3_access_key = ae_consts.ev(
            'PREDICTIONS_S3_ACCESS_KEY',
            ae_consts.S3_ACCESS_KEY)
        self.pt_s3_secret_key = ae_consts.ev(
            'PREDICTIONS_S3_SECRET_KEY',
            ae_consts.S3_SECRET_KEY)
        self.pt_s3_address = ae_consts.ev(
            'PREDICTIONS_S3_ADDRESS',
            ae_consts.S3_ADDRESS)
        self.pt_s3_secure = str(ae_consts.ev(
            'PREDICTIONS_S3_SECURE',
            ae_consts.S3_SECURE)) == '1'
        self.pt_s3_region = ae_consts.ev(
            'PREDICTIONS_S3_REGION_NAME',
            ae_consts.S3_REGION_NAME)
        self.pt_s3_bucket = self.algo_predictions_loc.split('/')[-2]
        self.pt_s3_key = self.algo_predictions_loc.split('/')[-1]

        self.config_dict = None
        self.use_config_file = algo_config
        if not self.use_config_file:
            if os.path.exists('./cfg/default_algo.json'):
                self.use_config_file = (
                    './cfg/default_algo.json')
            elif os.path.exists('/opt/sa/cfg/default_algo.json'):
                self.use_config_file = (
                    '/opt/sa/cfg/default_algo.json')
            else:
                log.critical(
                    f'Failed: missing algo_config argument pointing to a '
                    f'config file')
                return
        self.config_dict = json.loads(open(self.use_config_file, 'r').read())
        self.algo_mod_path = self.config_dict.get(
            'algo_path',
            ae_consts.ALGO_MODULE_PATH)
        if not os.path.exists(self.algo_mod_path):
            log.critical(
                f'missing algorithm module file from config: '
                f'{self.algo_mod_path}')
            return

        self.balance = 10000.0
        self.commission = 6.0
        self.use_start_date = None
        self.use_end_date = None

        self.inspect_datasets = False
        self.history_json_file = None
        self.run_this_date = None
        self.algo_obj = None
        self.algo_report_loc = None
        self.algo_extract_loc = None
        self.backtest_loc = None
        self.raise_on_err = True

        self.ssl_options = ae_consts.SSL_OPTIONS
        self.transport_options = ae_consts.TRANSPORT_OPTIONS
        self.broker_url = ae_consts.WORKER_BROKER_URL
        self.backend_url = ae_consts.WORKER_BACKEND_URL
        self.path_to_config_module = ae_consts.WORKER_CELERY_CONFIG_MODULE
        self.include_tasks = ae_consts.INCLUDE_TASKS
        self.load_from_s3_bucket = None
        self.load_from_s3_key = None
        self.load_from_redis_key = None
        self.load_from_file = None
        self.load_compress = True
        self.load_publish = True
        self.load_config = None
        self.report_redis_key = None
        self.report_s3_bucket = None
        self.report_s3_key = None
        self.report_file = None
        self.report_compress = True
        self.report_publish = False
        self.report_config = None
        self.history_redis_key = None
        self.history_s3_bucket = None
        self.history_s3_key = None
        self.history_file = None
        self.history_compress = True
        self.history_publish = True
        self.history_config = None
        self.extract_redis_key = None
        self.extract_s3_bucket = None
        self.extract_s3_key = None
        self.extract_file = None
        self.extract_save_dir = None
        self.extract_compress = False
        self.extract_publish = False
        self.extract_config = None
        self.s3_enabled = True
        self.s3_access_key = ae_consts.S3_ACCESS_KEY
        self.s3_secret_key = ae_consts.S3_SECRET_KEY
        self.s3_region_name = ae_consts.S3_REGION_NAME
        self.s3_address = ae_consts.S3_ADDRESS
        self.s3_bucket = ae_consts.S3_BUCKET
        self.s3_key = None
        self.s3_secure = ae_consts.S3_SECURE
        self.redis_enabled = True
        self.redis_address = ae_consts.REDIS_ADDRESS
        self.redis_key = None
        self.redis_password = ae_consts.REDIS_PASSWORD
        self.redis_db = ae_consts.REDIS_DB
        self.redis_expire = ae_consts.REDIS_EXPIRE
        self.redis_serializer = 'json'
        self.redis_encoding = 'utf-8'
        self.publish_to_s3 = True
        self.publish_to_redis = True
        self.publish_to_slack = False
        self.slack_enabled = False
        self.slack_code_block = False
        self.slack_full_width = False

        self.dataset_type = ae_consts.SA_DATASET_TYPE_ALGO_READY
        self.serialize_datasets = ae_consts.DEFAULT_SERIALIZED_DATASETS
        self.compress = False
        self.encoding = 'utf-8'
        self.debug = False
        self.num_rows = 0

        self.auto_fill = True
        self.timeseries = 'minute'
        self.trade_strategy = 'count'
        self.algo_history_s3_bucket = None
        self.algo_history_s3_key = None

        if self.start_date:
            self.use_start_date = f'{str(self.start_date)} 00:00:00'
            datetime.datetime.strptime(
                self.start_date,
                ae_consts.COMMON_DATE_FORMAT)
        if self.end_date:
            self.use_end_date = f'{str(self.end_date)} 00:00:00'
            datetime.datetime.strptime(
                self.end_date,
                ae_consts.COMMON_DATE_FORMAT)

        """
        Finalize the algo config
        """
        self.debug = False
        self.verbose_algo = verbose_algo
        self.verbose_processor = verbose_processor
        self.verbose_indicators = verbose_indicators

        if self.config_dict:
            if self.ticker:
                self.config_dict['ticker'] = self.ticker
            else:
                self.ticker = (
                    str(self.config_dict['ticker']).upper())
            self.config_dict['balance'] = self.balance
            self.config_dict['commission'] = self.commission

        if self.verbose_algo:
            self.config_dict['verbose'] = self.verbose_algo
        if self.verbose_processor:
            self.config_dict['verbose_processor'] = self.verbose_processor
        if self.verbose_indicators:
            self.config_dict['verbose_indicators'] = self.verbose_indicators
        if self.inspect_datasets:
            self.config_dict['inspect_datasets'] = self.inspect_datasets
        if self.run_this_date:
            self.config_dict['run_this_date'] = self.run_this_date

        if self.backtest_loc:
            if 's3://' in self.backtest_loc:
                self.load_from_s3_bucket = self.backtest_loc.split('/')[-2]
                self.load_from_s3_key = self.backtest_loc.split('/')[-1]
            elif 'redis://' in self.backtest_loc:
                self.load_from_redis_key = self.backtest_loc.split('/')[-1]
            elif 'file:/' in self.backtest_loc:
                self.load_from_file = self.backtest_loc.split(':')[-1]
            else:
                log.error(
                    'invalid -b <backtest dataset file> specified. '
                    f'{self.backtest_loc} '
                    'please use either: '
                    '-b file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-b s3://algoready/SPY-latest.json or '
                    '-b redis://SPY-latest')
            self.load_publish = True

        if self.algo_history_loc:
            if 's3://' in self.algo_history_loc:
                self.history_s3_bucket = self.algo_history_loc.split('/')[-2]
                self.history_s3_key = self.algo_history_loc.split('/')[-1]
            elif 'redis://' in self.algo_history_loc:
                self.history_redis_key = self.algo_history_loc.split('/')[-1]
            elif 'file:/' in self.algo_history_loc:
                self.history_file = self.algo_history_loc.split(':')[-1]
            else:
                log.error(
                    'invalid -p <backtest dataset file> specified. '
                    f'{self.algo_history_loc} '
                    'please use either: '
                    '-p file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-p s3://algoready/SPY-latest.json or '
                    '-p redis://SPY-latest')
            self.history_publish = True

        if self.algo_report_loc:
            if 's3://' in self.algo_report_loc:
                self.report_s3_bucket = self.algo_report_loc.split('/')[-2]
                self.report_s3_key = self.algo_report_loc.split('/')[-1]
            elif 'redis://' in self.algo_report_loc:
                self.report_redis_key = self.algo_report_loc.split('/')[-1]
            elif 'file:/' in self.algo_report_loc:
                self.report_file = self.algo_report_loc.split(':')[-1]
            else:
                log.error(
                    'invalid -o <backtest dataset file> specified. '
                    f'{self.algo_report_loc} '
                    'please use either: '
                    '-o file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-o s3://algoready/SPY-latest.json or '
                    '-o redis://SPY-latest')
            self.report_publish = True

        if self.algo_extract_loc:
            if 's3://' in self.algo_extract_loc:
                self.extract_s3_bucket = self.algo_extract_loc.split('/')[-2]
                self.extract_s3_key = self.algo_extract_loc.split('/')[-1]
            elif 'redis://' in self.algo_extract_loc:
                self.extract_redis_key = self.algo_extract_loc.split('/')[-1]
            elif 'file:/' in self.algo_extract_loc:
                self.extract_file = self.algo_extract_loc.split(':')[-1]
            else:
                log.error(
                    'invalid -e <backtest dataset file> specified. '
                    f'{self.algo_extract_loc} '
                    'please use either: '
                    '-e file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-e s3://algoready/SPY-latest.json or '
                    '-e redis://SPY-latest')
            self.extract_publish = True

        self.use_name = self.config_dict.get(
            'name',
            'missing-algo-name')
        self.auto_fill = self.config_dict.get(
            'auto_fill',
            self.auto_fill)
        self.timeseries = self.config_dict.get(
            'timeseries',
            self.timeseries)
        self.trade_strategy = self.config_dict.get(
            'trade_strategy',
            self.trade_strategy)

        self.iex_datasets = ae_consts.IEX_DATASETS_DEFAULT
        self.common_fetch_vals = {}
        self.common_fetch_vals['ticker'] = self.ticker
        self.common_fetch_vals['celery_disabled'] = True
        self.common_fetch_vals['label'] = 'lp'
        self.common_fetch_vals['iex_datasets'] = self.iex_datasets
        self.common_fetch_vals['s3_enabled'] = False
        self.common_fetch_vals['s3_address'] = None
        self.common_fetch_vals['s3_bucket'] = None
        self.common_fetch_vals['s3_key'] = None
        self.common_fetch_vals['s3_access_key'] = None
        self.common_fetch_vals['s3_secret_key'] = None
        self.common_fetch_vals['s3_region_name'] = None
        self.common_fetch_vals['s3_secure'] = None
        self.common_fetch_vals['redis_enabled'] = True
        self.common_fetch_vals['redis_address'] = self.redis_address
        self.common_fetch_vals['redis_password'] = self.redis_password
        self.common_fetch_vals['redis_expire'] = None
        self.common_fetch_vals['redis_db'] = self.redis_db

        self.history_df = None
        self.slack_enabled = False
    # end of __init__

    def start(
            self):
        """start

        Start the algorithm backtest
        """

        log.info(
            'starting algo '
            f's3://{self.history_s3_bucket}/{self.history_s3_key}')

        self.algo_res = run_custom_algo.run_custom_algo(
            mod_path=self.algo_mod_path,
            ticker=self.config_dict['ticker'],
            balance=self.config_dict['balance'],
            commission=self.config_dict['commission'],
            name=self.use_name,
            start_date=self.use_start_date,
            end_date=self.use_end_date,
            auto_fill=self.auto_fill,
            config_dict=self.config_dict,
            load_from_s3_bucket=self.load_from_s3_bucket,
            load_from_s3_key=self.load_from_s3_key,
            load_from_redis_key=self.load_from_redis_key,
            load_from_file=self.load_from_file,
            load_compress=self.load_compress,
            load_publish=self.load_publish,
            load_config=self.load_config,
            report_redis_key=self.report_redis_key,
            report_s3_bucket=self.report_s3_bucket,
            report_s3_key=self.report_s3_key,
            report_file=self.report_file,
            report_compress=self.report_compress,
            report_publish=self.report_publish,
            report_config=self.report_config,
            history_redis_key=self.history_redis_key,
            history_s3_bucket=self.history_s3_bucket,
            history_s3_key=self.history_s3_key,
            history_file=self.history_file,
            history_compress=self.history_compress,
            history_publish=self.history_publish,
            history_config=self.history_config,
            extract_redis_key=self.extract_redis_key,
            extract_s3_bucket=self.extract_s3_bucket,
            extract_s3_key=self.extract_s3_key,
            extract_file=self.extract_file,
            extract_save_dir=self.extract_save_dir,
            extract_compress=self.extract_compress,
            extract_publish=self.extract_publish,
            extract_config=self.extract_config,
            publish_to_slack=self.publish_to_slack,
            publish_to_s3=self.publish_to_s3,
            publish_to_redis=self.publish_to_redis,
            dataset_type=self.dataset_type,
            serialize_datasets=self.serialize_datasets,
            compress=self.compress,
            encoding=self.encoding,
            redis_enabled=self.redis_enabled,
            redis_key=self.redis_key,
            redis_address=self.redis_address,
            redis_db=self.redis_db,
            redis_password=self.redis_password,
            redis_expire=self.redis_expire,
            redis_serializer=self.redis_serializer,
            redis_encoding=self.redis_encoding,
            s3_enabled=self.s3_enabled,
            s3_key=self.s3_key,
            s3_address=self.s3_address,
            s3_bucket=self.s3_bucket,
            s3_access_key=self.s3_access_key,
            s3_secret_key=self.s3_secret_key,
            s3_region_name=self.s3_region_name,
            s3_secure=self.s3_secure,
            slack_enabled=self.slack_enabled,
            slack_code_block=self.slack_code_block,
            slack_full_width=self.slack_full_width,
            dataset_publish_extract=self.extract_publish,
            dataset_publish_history=self.history_publish,
            dataset_publish_report=self.report_publish,
            run_on_engine=self.run_on_engine,
            auth_url=self.broker_url,
            backend_url=self.backend_url,
            include_tasks=self.include_tasks,
            ssl_options=self.ssl_options,
            transport_options=self.transport_options,
            path_to_config_module=self.path_to_config_module,
            timeseries=self.timeseries,
            trade_strategy=self.trade_strategy,
            verbose=self.verbose_algo,
            raise_on_err=self.raise_on_err)

        self.wait_for_algo_to_finish()
    # end of start

    def wait_for_algo_to_finish(
            self):
        """wait_for_algo_to_finish

        wait until the algorithm finishes
        """

        self.algo_history_s3_bucket = None
        self.algo_history_s3_key = None
        self.task_id = None
        if self.run_on_engine:
            self.task_id = self.algo_res.get(
                'rec', {}).get('task_id', None)
        if self.task_id:
            log.info(
                f'waiting - algo task_id={self.task_id} to finish')
            res = self.task_id.get()
            history_config = res.get(
                'algo_req', {}).get(
                    'history_config', None)
            self.algo_history_s3_bucket = history_config.get(
                's3_bucket', None)
            self.algo_history_s3_key = history_config.get(
                's3_key', None)
            self.load_trading_history()
        else:
            history_records = self.algo_res.get(
                'rec', {}).get(
                    'history', None)
            self.algo_history_s3_bucket = self.history_s3_bucket
            self.algo_history_s3_key = self.history_s3_key
            self.history_df = pd.DataFrame(history_records)
            self.determine_latest_times_in_history()

        self.num_rows = len(self.history_df.index)

        log.info(
            f'loaded history rows={self.num_rows} '
            f'dates={self.start_date} to {self.end_date}')
    # end of wait_for_algo_to_finish

    def determine_latest_times_in_history(
            self):
        """determine_latest_times_in_history

        determine the latest minute or day in the pricing dataset
        and convert ``date`` and ``minute`` columns to ``datetime``
        objects
        """
        self.history_df['date'] = pd.to_datetime(
            self.history_df['date'])
        self.start_day = self.history_df['date'].iloc[0]
        self.end_day = self.history_df['date'].iloc[-1]
        self.start_date = self.history_df['date'].iloc[0]
        self.end_date = self.history_df['date'].iloc[-1]
        if 'minute' in self.history_df:
            self.history_df['minute'] = pd.to_datetime(
                self.history_df['minute'])
            self.start_date = self.history_df['minute'].iloc[0]
            self.end_date = self.history_df['minute'].iloc[-1]
    # end of determine_latest_times_in_history

    def load_trading_history(
            self,
            s3_access_key=None,
            s3_secret_key=None,
            s3_address=None,
            s3_region=None,
            s3_bucket=None,
            s3_key=None,
            s3_secure=ae_consts.NOT_SET,
            **kwargs):
        """load_trading_history

        Helper for loading an algorithm ``Trading History`` from
        S3

        :param s3_access_key: access key
        :param s3_secret_key: secret
        :param s3_address: address
        :param s3_region: region
        :param s3_bucket: bucket
        :param s3_key: key
        :param s3_secure: secure flag
        :param kwargs: support for keyword arg dict
        """
        use_s3_access_key = self.s3_access_key
        use_s3_secret_key = self.s3_secret_key
        use_s3_address = self.s3_address
        use_s3_region = self.s3_region_name
        use_s3_bucket = self.s3_bucket
        use_s3_key = self.s3_key
        use_s3_secure = self.s3_secure

        if s3_access_key:
            use_s3_access_key = s3_access_key
        if s3_secret_key:
            use_s3_secret_key = s3_secret_key
        if s3_address:
            use_s3_address = s3_address
        if s3_region:
            use_s3_region = s3_region
        if s3_bucket:
            use_s3_bucket = s3_bucket
        if s3_key:
            use_s3_key = s3_key
        if s3_secure != ae_consts.NOT_SET:
            use_s3_secure = s3_secure

        if s3_key and s3_bucket:
            log.info(
                f'using - td s3://{use_s3_address}/'
                f'{use_s3_bucket}/{use_s3_key}')
        else:
            log.info(
                f'load - td s3://{use_s3_address}/'
                f'{use_s3_bucket}/{use_s3_key}')

        load_res = load_history_utils.load_history_dataset(
            s3_access_key=use_s3_access_key,
            s3_secret_key=use_s3_secret_key,
            s3_address=use_s3_address,
            s3_region_name=use_s3_region,
            s3_bucket=use_s3_bucket,
            s3_key=use_s3_key,
            s3_secure=use_s3_secure)

        if self.ticker not in load_res:
            log.critical(
                'failed to load history: '
                f's3://{self.s3_bucket}/{self.s3_key}')
            self.history_df = None
            return

        self.history_df = load_res[self.ticker]
        self.determine_latest_times_in_history()

        log.info(
            f'found - td s3://{use_s3_address}/'
            f'{use_s3_bucket}/{use_s3_key} '
            f'rows={len(self.history_df.index)}')
    # end of load_trading_history

    def get_history(
            self):
        """get_history"""
        return self.history_df
    # end of get_history

    def get_latest_day(
            self):
        """get_latest_day"""
        if 'date' in self.history_df:
            return self.end_day
        else:
            return None
    # end of get_latest_day

    def get_latest_minute(
            self):
        """get_latest_minute"""
        if 'minute' in self.history_df:
            return self.end_date
        else:
            return None
    # end of get_latest_minute

    def publish_trading_history(
            self,
            records_for_history,
            pt_s3_access_key=None,
            pt_s3_secret_key=None,
            pt_s3_address=None,
            pt_s3_region=None,
            pt_s3_bucket=None,
            pt_s3_key=None,
            pt_s3_secure=ae_consts.NOT_SET,
            **kwargs):
        """publish_trading_history

        Helper for publishing a trading history
        to another S3 service like AWS

        :param records_for_history: list of dictionaries
            for the history file
        :param pt_s3_access_key: access key
        :param pt_s3_secret_key: secret
        :param pt_s3_address: address
        :param pt_s3_region: region
        :param pt_s3_bucket: bucket
        :param pt_s3_key: key
        :param pt_s3_secure: secure flag
        :param kwargs: support for keyword arg dict
        """
        use_s3_access_key = self.pt_s3_access_key
        use_s3_secret_key = self.pt_s3_secret_key
        use_s3_address = self.pt_s3_address
        use_s3_region = self.pt_s3_region
        use_s3_bucket = self.pt_s3_bucket
        use_s3_key = self.pt_s3_key
        use_s3_secure = self.pt_s3_secure

        use_s3_enabled = kwargs.get(
            's3_enabled', True)
        use_redis_enabled = kwargs.get(
            'redis_enabled', False)
        use_redis_address = kwargs.get(
            'redis_address', None)
        use_redis_db = kwargs.get(
            'redis_db', None)
        use_redis_key = kwargs.get(
            'redis_key', None)
        use_redis_password = kwargs.get(
            'redis_password', None)
        use_redis_expire = kwargs.get(
            'redis_expire', None)
        use_redis_serializer = kwargs.get(
            'redis_serializer', 'json')
        use_redis_encoding = kwargs.get(
            'redis_encoding', 'utf-8')
        verbose = kwargs.get(
            'verbose', False)

        if pt_s3_access_key:
            use_s3_access_key = pt_s3_access_key
        if pt_s3_secret_key:
            use_s3_secret_key = pt_s3_secret_key
        if pt_s3_address:
            use_s3_address = pt_s3_address
        if pt_s3_region:
            use_s3_region = pt_s3_region
        if pt_s3_bucket:
            use_s3_bucket = pt_s3_bucket
        if pt_s3_key:
            use_s3_key = pt_s3_key
        if pt_s3_secure != ae_consts.NOT_SET:
            use_s3_secure = pt_s3_secure

        rec = {
            'tickers': self.ticker,
            'version': int(ae_consts.ALGO_HISTORY_VERSION),
            'last_trade_date': ae_utils.get_last_close_str(),
            'algo_config_dict': self.config_dict,
            'algo_name':  self.use_name,
            'created': ae_utils.utc_now_str(),
            self.ticker: records_for_history
        }

        num_bytes = len(str(rec))
        num_mb = ae_consts.get_mb(num_bytes)

        msg = (
            f'publish - {self.ticker} - {rec["last_trade_date"]} '
            # f'{use_s3_access_key} with: {use_s3_secret_key} '
            f's3_loc={use_s3_address}/{use_s3_bucket}/{use_s3_key} '
            f'mb={num_mb}MB')
        log.info(msg)

        publish.publish(
            data=rec,
            label='pub',
            df_compress=True,
            compress=False,
            convert_to_dict=False,
            output_file=None,
            redis_enabled=use_redis_enabled,
            redis_key=use_redis_key,
            redis_address=use_redis_address,
            redis_db=use_redis_db,
            redis_password=use_redis_password,
            redis_expire=use_redis_expire,
            redis_serializer=use_redis_serializer,
            redis_encoding=use_redis_encoding,
            s3_enabled=use_s3_enabled,
            s3_key=use_s3_key,
            s3_address=use_s3_address,
            s3_bucket=use_s3_bucket,
            s3_access_key=use_s3_access_key,
            s3_secret_key=use_s3_secret_key,
            s3_region_name=use_s3_region,
            s3_secure=use_s3_secure,
            slack_enabled=False,
            verbose=verbose)
    # end of publish_trading_history

    def latest(
            self,
            date_str=None,
            start_row=-200,
            extract_iex=True,
            extract_yahoo=False,
            extract_td=True,
            verbose=False,
            **kwargs):
        """latest

        Run the algorithm with the latest pricing data. Also
        supports running a backtest for a historical date in
        the pricing history (format ``YYYY-MM-DD``)

        :param date_str: optional - string start date ``YYYY-MM-DD``
            default is the latest close date
        :param start_row: negative number of rows back
            from the end of the list in the data
            default is ``-200`` where this means the algorithm
            will process the latest 200 rows in the minute
            dataset
        :param extract_iex: bool flag for extracting from ``IEX``
        :param extract_yahoo: bool flag for extracting from ``Yahoo``
            which is disabled as of 1/2019
        :param extract_td: bool flag for extracting from ``Tradier``
        :param verbose: bool flag for logs
        :param kwargs: keyword arg dict
        """
        use_date_str = date_str
        if not use_date_str:
            use_date_str = ae_utils.get_last_close_str()

        log.info(
            f'creating algo')
        self.algo_obj = base_algo.BaseAlgo(
            ticker=self.config_dict['ticker'],
            balance=self.config_dict['balance'],
            commission=self.config_dict['commission'],
            name=self.use_name,
            start_date=self.use_start_date,
            end_date=self.use_end_date,
            auto_fill=self.auto_fill,
            config_dict=self.config_dict,
            load_from_s3_bucket=self.load_from_s3_bucket,
            load_from_s3_key=self.load_from_s3_key,
            load_from_redis_key=self.load_from_redis_key,
            load_from_file=self.load_from_file,
            load_compress=self.load_compress,
            load_publish=self.load_publish,
            load_config=self.load_config,
            report_redis_key=self.report_redis_key,
            report_s3_bucket=self.report_s3_bucket,
            report_s3_key=self.report_s3_key,
            report_file=self.report_file,
            report_compress=self.report_compress,
            report_publish=self.report_publish,
            report_config=self.report_config,
            history_redis_key=self.history_redis_key,
            history_s3_bucket=self.history_s3_bucket,
            history_s3_key=self.history_s3_key,
            history_file=self.history_file,
            history_compress=self.history_compress,
            history_publish=self.history_publish,
            history_config=self.history_config,
            extract_redis_key=self.extract_redis_key,
            extract_s3_bucket=self.extract_s3_bucket,
            extract_s3_key=self.extract_s3_key,
            extract_file=self.extract_file,
            extract_save_dir=self.extract_save_dir,
            extract_compress=self.extract_compress,
            extract_publish=self.extract_publish,
            extract_config=self.extract_config,
            publish_to_slack=self.publish_to_slack,
            publish_to_s3=self.publish_to_s3,
            publish_to_redis=self.publish_to_redis,
            dataset_type=self.dataset_type,
            serialize_datasets=self.serialize_datasets,
            compress=self.compress,
            encoding=self.encoding,
            redis_enabled=self.redis_enabled,
            redis_key=self.redis_key,
            redis_address=self.redis_address,
            redis_db=self.redis_db,
            redis_password=self.redis_password,
            redis_expire=self.redis_expire,
            redis_serializer=self.redis_serializer,
            redis_encoding=self.redis_encoding,
            s3_enabled=self.s3_enabled,
            s3_key=self.s3_key,
            s3_address=self.s3_address,
            s3_bucket=self.s3_bucket,
            s3_access_key=self.s3_access_key,
            s3_secret_key=self.s3_secret_key,
            s3_region_name=self.s3_region_name,
            s3_secure=self.s3_secure,
            slack_enabled=self.slack_enabled,
            slack_code_block=self.slack_code_block,
            slack_full_width=self.slack_full_width,
            dataset_publish_extract=self.extract_publish,
            dataset_publish_history=self.history_publish,
            dataset_publish_report=self.report_publish,
            run_on_engine=self.run_on_engine,
            auth_url=self.broker_url,
            backend_url=self.backend_url,
            include_tasks=self.include_tasks,
            ssl_options=self.ssl_options,
            transport_options=self.transport_options,
            path_to_config_module=self.path_to_config_module,
            timeseries=self.timeseries,
            trade_strategy=self.trade_strategy,
            verbose=False,
            raise_on_err=self.raise_on_err)

        log.info(
            f'run latest - start')

        ticker = self.config_dict['ticker']
        dataset_id = f'{ticker}_{use_date_str}'
        self.common_fetch_vals['base_key'] = dataset_id
        verbose_extract = self.config_dict.get('verbose_extract', False)
        indicator_datasets = self.algo_obj.get_indicator_datasets()
        ticker_data = build_dataset_node.build_dataset_node(
            ticker=ticker,
            date=use_date_str,
            datasets=indicator_datasets,
            service_dict=self.common_fetch_vals,
            verbose=verbose_extract)

        algo_data_req = {
            ticker: [
                {
                    'id': dataset_id,  # id is currently the cache key in redis
                    'date': use_date_str,  # used to confirm dates in asc order
                    'data': ticker_data,
                    'start_row': start_row
                }
            ]
        }

        if verbose:
            log.info(
                f'extract - {dataset_id} '
                f'dataset={len(algo_data_req[ticker])}')

        # this could be a separate celery task
        try:
            if verbose:
                log.info(
                    f'handle_data START - {dataset_id}')
            self.algo_obj.handle_data(
                data=algo_data_req)
            if verbose:
                log.info(
                    f'handle_data END - {dataset_id}')
        except Exception as e:
            a_name = self.algo_obj.get_name()
            a_debug_msg = self.algo_obj.get_debug_msg()
            if not a_debug_msg:
                a_debug_msg = 'debug message not set'
            # a_config_dict = ae_consts.ppj(self.algo_obj.config_dict)
            msg = (
                f'{dataset_id} - algo={a_name} '
                f'encountered exception in handle_data tickers={ticker} '
                f'from ex={e} '
                f'and failed during operation: {a_debug_msg}')
            log.critical(f'{msg}')
        # end try/ex

        log.info('run latest - create history')

        history_ds = self.algo_obj.create_history_dataset()
        self.history_df = pd.DataFrame(history_ds[ticker])
        self.determine_latest_times_in_history()

        self.num_rows = len(self.history_df.index)

        if verbose:
            log.info(self.history_df[['minute', 'close']].tail(5))

        log.info(
            f'run latest minute={self.end_date} - '
            f'rows={self.num_rows} - done')

        return self.get_history()
    # end of latest

# end of AlgoRunner



#!/usr/bin/env python

"""
A tool for plotting an algorithm's ``Trading History`` from
a locally saved file from running the backtester with
the save to file option enabled:

::

    run_backtest_and_plot_history.py -t SPY -f <SAVE_HISTORY_TO_THIS_FILE>
"""

import os
import argparse
import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.plot_trading_history as plot_trading_history
import spylunking.log.setup_logging as log_utils


log = log_utils.build_colorized_logger(
    name='plot-history',
    log_config_path=ae_consts.LOG_CONFIG_PATH)


def plot_local_history_file():
    """plot_local_history_file

    Run a derived algorithm with an algorithm config dictionary

    :param config_dict: algorithm config dictionary
    """

    log.debug('start - plot')

    parser = argparse.ArgumentParser(
        description=(
            'plot a local algorithm trading history file'))
    parser.add_argument(
        '-f',
        help=(
            'plot this trading history dataframe '
            'saved in this file'),
        required=False,
        dest='history_json_file')
    parser.add_argument(
        '-d',
        help=(
            'debug'),
        required=False,
        dest='debug',
        action='store_true')
    args = parser.parse_args()

    history_json_file = None
    debug = False

    if args.history_json_file:
        history_json_file = args.history_json_file
    if args.debug:
        debug = True

    if not history_json_file:
        log.error(
            'usage error - please run with: '
            '-f <path to local trading history file>')
        return
    elif not os.path.exists(history_json_file):
        log.error(f'did not find trading history file={history_json_file}')
        return
    # end of checking the file arg is set and exists on disk

    log.info(f'plotting history to: {history_json_file}')
    history_df = pd.read_json(
        history_json_file,
        orient='records')

    history_df['date'] = pd.to_datetime(
        history_df['date'])
    history_df['minute'] = pd.to_datetime(
        history_df['minute'])
    ticker = history_df['ticker'].iloc[0]

    log.info('plotting history')

    first_date = history_df['date'].iloc[0]
    end_date = history_df['date'].iloc[-1]
    title = (
        f'Trading History {ticker}\n'
        f'Backtest dates from {first_date} to {end_date}')
    use_xcol = 'date'
    use_as_date_format = '%d\n%b'
    use_minute = False
    if 'minute' in history_df:
        found_valid_minute = history_df['minute'].iloc[0]
        if found_valid_minute:
            use_minute = True

    if use_minute:
        use_xcol = 'minute'
        use_as_date_format = '%d %H:%M:%S\n%b'
    xlabel = 'Dates vs Algo values'
    ylabel = 'Algo values'
    df_filter = (history_df['close'] > 1.00)

    # set default hloc columns:
    blue = None
    green = None
    orange = None

    red = 'close'
    blue = 'balance'

    if debug:
        for i, r in history_df.iterrows():
            log.info(f'{r["minute"]} - {r["close"]}')
    # end of debug

    show_plot = True
    if show_plot:
        plot_trading_history.plot_trading_history(
            title=title,
            df=history_df,
            red=red,
            blue=blue,
            green=green,
            orange=orange,
            date_col=use_xcol,
            date_format=use_as_date_format,
            xlabel=xlabel,
            ylabel=ylabel,
            df_filter=df_filter,
            show_plot=True,
            dropna_for_all=True)

# end of plot_local_history_file


if __name__ == '__main__':
    plot_local_history_file()



"""
Mock boto3 s3 objects
"""

import os
import json
import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def mock_s3_read_contents_from_key_ev(
        s3,
        s3_bucket_name,
        s3_key,
        encoding,
        convert_as_json):
    """mock_s3_read_contents_from_key

    :param s3: s3 client
    :param s3_bucket_name: bucket name
    :param s3_key: key
    :param encoding: utf-8
    :param convert_as_json: convert to json
    """

    env_key = 'TEST_S3_CONTENTS'
    str_contents = ae_consts.ev(
        env_key,
        None)

    log.info(
        f'returning mock s3={s3_bucket_name}:{s3_key} contents={str_contents} '
        f'encoding={encoding} json={convert_as_json} env_key={env_key}')

    if not str_contents:
        return str_contents

    if convert_as_json:
        return json.loads(str_contents)
# end of mock_s3_read_contents_from_key_ev


def mock_publish_from_s3_to_redis(
        work_dict):
    """mock_publish_from_s3_to_redis

    :param work_dict: dictionary for driving the task
    """

    env_key = 'TEST_S3_CONTENTS'
    redis_key = work_dict.get(
        'redis_key',
        env_key)
    str_dict = ae_consts.ev(
        env_key,
        None)
    log.info(
        'mock_publish_from_s3_to_redis - '
        f'setting key={redis_key} value={str_dict}')
    data = None
    if str_dict:
        os.environ[redis_key] = str_dict
        data = str_dict.encode('utf-8')
    else:
        os.environ[redis_key] = ''
        data = None

    status = ae_consts.SUCCESS
    err = None
    return {
        'status': status,
        'err': err,
        'rec': {
            'data': data
        }
    }
# end of mock_publish_from_s3_to_redis


def mock_publish_from_s3_to_redis_err(
        work_dict):
    """mock_publish_from_s3_to_redis_err

    :param work_dict: dictionary for driving the task
    """

    env_key = 'TEST_S3_CONTENTS'
    redis_key = work_dict.get(
        'redis_key',
        env_key)
    str_dict = ae_consts.ev(
        env_key,
        None)
    log.info(
        'mock_publish_from_s3_to_redis_err - '
        f'setting key={redis_key} value={str_dict}')
    data = None
    if str_dict:
        os.environ[redis_key] = str_dict
        data = str_dict.encode('utf-8')
    else:
        os.environ[redis_key] = ''
        data = None

    status = ae_consts.ERR
    err = None
    return {
        'status': status,
        'err': err,
        'rec': {
            'data': data
        }
    }
# end of mock_publish_from_s3_to_redis_err


def mock_publish_from_s3_exception(
        work_dict):
    """mock_publish_from_s3_exception

    :param work_dict: dictionary for driving the task
    """
    raise Exception(
        'test mock_publish_from_s3_exception')
# end of mock_publish_from_s3_exception


class MockBotoS3Bucket:
    """MockBotoS3Bucket"""

    def __init__(
            self,
            name):
        """__init__

        build mock bucket

        :param name: name of the bucket
        """
        self.name = name
        self.datas = []  # payloads uploaded to s3
        self.keys = []   # keys uploaded to s3
    # end of __init__

    def put_object(
            self,
            Key=None,
            Body=None):
        """put_object

        :param Key: new Key name
        :param Body: new Payload in Key
        """

        log.debug(
            f'mock - MockBotoS3Bucket.put_object(Key={Key}, '
            f'Body={Body})')

        self.keys.append(Key)
        self.datas.append(Body)
    # end of put_object

# end of MockBotoS3Bucket


class MockBotoS3AllBuckets:
    """MockBotoS3AllBuckets"""

    def __init__(
            self):
        """__init__"""
        self.buckets = {}
    # end of __init__

    def add(
            self,
            bucket_name):
        """add

        :param bucket_name: bucket name to add
        """
        if bucket_name not in self.buckets:
            log.info(
                f'adding bucket={bucket_name} total={len(self.buckets) + 1}')
            self.buckets[bucket_name] = MockBotoS3Bucket(
                name=bucket_name)

        return self.buckets[bucket_name]
    # end of add

    def all(
            self):
        """all"""
        return self.buckets
    # end of all

# end of MockBotoS3AllBuckets


class MockBotoS3:
    """MockBotoS3"""

    def __init__(
            self,
            name='mock_s3',
            endpoint_url=None,
            aws_access_key_id=None,
            aws_secret_access_key=None,
            region_name=None,
            config=None):
        """__init__

        build mock object

        :param name: name of client
        :param endpoint_url: endpoint url
        :param aws_access_key_id: aws access key
        :param aws_secret_access_key: aws secret key
        :param region_name: region name
        :param config: config object
        """

        self.name = name
        self.endpoint_url = endpoint_url
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.region_name = region_name
        self.config = config
        self.buckets = MockBotoS3AllBuckets()
        self.keys = []
    # end of __init__

    def Bucket(
            self,
            name):
        """Bucket

        :param name: name of new bucket
        """
        log.info(
            f'MockBotoS3.Bucket({name})')
        return self.buckets.add(
            bucket_name=name)
    # end of Bucket

    def create_bucket(
            self,
            Bucket=None):
        """create_bucket

        :param bucket_name: name of the new bucket
        """
        log.info(
            f'mock - MockBotoS3.create_bucket(Bucket={Bucket})')
        return self.buckets.add(
            bucket_name=Bucket)
    # end of create_bucket

# end of MockBotoS3


def build_boto3_resource(
        name='mock_s3',
        endpoint_url=None,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        region_name=None,
        config=None):
    """build_boto3_resource

    :param name: name of client
    :param endpoint_url: endpoint url
    :param aws_access_key_id: aws access key
    :param aws_secret_access_key: aws secret key
    :param region_name: region name
    :param config: config object
    """

    if 's3' in name.lower():
        return MockBotoS3(
            name='mock_s3',
            endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
            config=config)
    else:
        return MockBotoS3(
            name='mock_s3',
            endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
            config=config)
# end of build_boto3_resource



"""
Extract an IEX dataset from Redis and
return it as a ``pandas.DataFrame`` or None
"""

import copy
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.api_requests as api_requests
import analysis_engine.iex.consts as iex_consts
import analysis_engine.extract_utils as extract_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)
keys = {
    'company': iex_consts.DATAFEED_COMPANY,
    'daily': iex_consts.DATAFEED_DAILY,
    'dividends': iex_consts.DATAFEED_DIVIDENDS,
    'earnings': iex_consts.DATAFEED_EARNINGS,
    'financials': iex_consts.DATAFEED_FINANCIALS,
    'minute': iex_consts.DATAFEED_MINUTE,
    'news1': iex_consts.DATAFEED_NEWS,
    'peers': iex_consts.DATAFEED_PEERS,
    'quote': iex_consts.DATAFEED_QUOTE,
    'stats': iex_consts.DATAFEED_STATS
}


def extract_daily_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_daily_dataset

    Fetch the IEX daily data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.iex.extract_df_from_redis as iex_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        daily_status, daily_df = iex_extract.extract_daily_dataset(
            ticker='SPY')
        print(daily_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    return extract_dataset(
        key='daily',
        work_dict=work_dict,
        ticker=ticker,
        date=date,
        scrub_mode=scrub_mode,
        verbose=verbose)
# end of extract_daily_dataset


def extract_minute_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_minute_dataset

    Fetch the IEX minute intraday data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.iex.extract_df_from_redis as iex_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        minute_status, minute_df = iex_extract.extract_minute_dataset(
            ticker='SPY')
        print(minute_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    return extract_dataset(
        key='minute',
        work_dict=work_dict,
        ticker=ticker,
        date=date,
        scrub_mode=scrub_mode,
        verbose=verbose)
# end of extract_minute_dataset


def extract_quote_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_quote_dataset

    Fetch the IEX quote data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.iex.extract_df_from_redis as iex_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        quote_status, quote_df = iex_extract.extract_quote_dataset(
            ticker='SPY')
        print(quote_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    return extract_dataset(
        key='quote',
        work_dict=work_dict,
        ticker=ticker,
        date=date,
        scrub_mode=scrub_mode,
        verbose=verbose)
# end of extract_quote_dataset


def extract_stats_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_stats_dataset

    Fetch the IEX statistics data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.iex.extract_df_from_redis as iex_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        stats_status, stats_df = iex_extract.extract_stats_dataset(
            ticker='SPY')
        print(stats_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    return extract_dataset(
        key='stats',
        work_dict=work_dict,
        ticker=ticker,
        date=date,
        scrub_mode=scrub_mode,
        verbose=verbose)
# end of extract_stats_dataset


def extract_peers_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_peers_dataset

    Fetch the IEX peers data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.iex.extract_df_from_redis as iex_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        peers_status, peers_df = iex_extract.extract_peers_dataset(
            ticker='SPY')
        print(peers_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    return extract_dataset(
        key='peers',
        work_dict=work_dict,
        ticker=ticker,
        date=date,
        scrub_mode=scrub_mode,
        verbose=verbose)
# end of extract_peers_dataset


def extract_news_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_news_dataset

    Fetch the IEX news data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.iex.extract_df_from_redis as iex_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        news_status, news_df = iex_extract.extract_news_dataset(
            ticker='SPY')
        print(news_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    return extract_dataset(
        key='news1',
        work_dict=work_dict,
        ticker=ticker,
        date=date,
        scrub_mode=scrub_mode,
        verbose=verbose)
# end of extract_news_dataset


def extract_financials_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_financials_dataset

    Fetch the IEX financial data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.iex.extract_df_from_redis as iex_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        fin_status, fin_df = iex_extract.extract_financials_dataset(
            ticker='SPY')
        print(fin_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    return extract_dataset(
        key='financials',
        work_dict=work_dict,
        ticker=ticker,
        date=date,
        scrub_mode=scrub_mode,
        verbose=verbose)
# end of extract_financials_dataset


def extract_earnings_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_earnings_dataset

    Fetch the IEX earnings data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.iex.extract_df_from_redis as iex_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        earn_status, earn_df = iex_extract.extract_earnings_dataset(
            ticker='SPY')
        print(earn_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    return extract_dataset(
        key='earnings',
        work_dict=work_dict,
        ticker=ticker,
        date=date,
        scrub_mode=scrub_mode,
        verbose=verbose)
# end of extract_earnings_dataset


def extract_dividends_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='sort-by-date',
        verbose=False):
    """extract_dividends_dataset

    Fetch the IEX dividends data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.iex.extract_df_from_redis as iex_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        div_status, div_df = iex_extract.extract_dividends_dataset(
            ticker='SPY')
        print(div_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    return extract_dataset(
        key='dividends',
        work_dict=work_dict,
        ticker=ticker,
        date=date,
        scrub_mode=scrub_mode,
        verbose=verbose)
# end of extract_dividends_dataset


def extract_company_dataset(
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='NO_SORT',
        verbose=False):
    """extract_company_dataset

    Fetch the IEX company data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    .. code-block:: python

        import analysis_engine.iex.extract_df_from_redis as iex_extract

        # extract by historical date is also supported as an arg
        # date='2019-02-15'
        comp_status, comp_df = iex_extract.extract_company_dataset(
            ticker='SPY')
        print(comp_df)

    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    return extract_dataset(
        key='company',
        work_dict=work_dict,
        ticker=ticker,
        date=date,
        scrub_mode=scrub_mode,
        verbose=verbose)
# end of extract_company_dataset


def extract_dataset(
        key,
        work_dict=None,
        ticker=None,
        date=None,
        scrub_mode='NO_SORT',
        verbose=False):
    """extract_dataset

    Fetch the IEX key data for a ticker and
    return a tuple (status, ``pandas.Dataframe``)

    :param key: IEX dataset key
    :param work_dict: dictionary of args
    :param ticker: optional - string ticker to extract
    :param date: optional - string date to extract
        formatted ``YYYY-MM-DD``
    :param scrub_mode: type of scrubbing handler to run
    :param verbose: optional - boolean for turning on logging
    """
    if not key or key not in keys:
        log.error(
            f'unsupported extract key={key} in keys={keys}')
        return None
    label = key
    df_type = keys[key]
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    latest_close_date = ae_utils.get_last_close_str()

    use_date = date
    if work_dict:
        if not ticker:
            ticker = work_dict.get('ticker', None)
    if not work_dict:
        work_dict = api_requests.get_ds_dict(
            ticker=ticker)

    req = copy.deepcopy(work_dict)

    if not use_date:
        use_date = latest_close_date

    redis_key = f'{ticker}_{use_date}_{key}'
    req['redis_key'] = redis_key
    req['s3_key'] = redis_key

    if verbose:
        log.info(
            f'{label} - {df_str} - '
            f'redis_key={req["redis_key"]} '
            f's3_key={req["s3_key"]} '
            f'{ae_consts.ppj(req)}')

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_dataset



#!/usr/bin/env python

import celery
import analysis_engine.consts as consts
import analysis_engine.work_tasks.get_celery_app as get_celery_app
import spylunking.log.setup_logging as log_utils


# Disable celery log hijacking
# https://github.com/celery/celery/issues/2509
@celery.signals.setup_logging.connect
def setup_celery_logging(**kwargs):
    pass


log = log_utils.build_colorized_logger(
    name=consts.APP_NAME,
    log_config_path=consts.LOG_CONFIG_PATH)

log.info(f'start - {consts.APP_NAME}')

log.info(
    f'broker={consts.WORKER_BROKER_URL} backend={consts.WORKER_BACKEND_URL} '
    f'config={consts.WORKER_CELERY_CONFIG_MODULE} '
    f'include_tasks={consts.WORKER_TASKS}')

# Get the Celery app from the project's get_celery_app module
app = get_celery_app.get_celery_app(
    name=consts.APP_NAME,
    path_to_config_module=consts.WORKER_CELERY_CONFIG_MODULE,
    auth_url=consts.WORKER_BROKER_URL,
    backend_url=consts.WORKER_BACKEND_URL,
    include_tasks=consts.INCLUDE_TASKS)

log.info('starting celery')
app.start()

log.info(f'end - {consts.APP_NAME}')



"""
Extract an Yahoo dataset from Redis (S3 support coming soon) and
load it into a ``pandas.DataFrame``

Supported environment variables:

::

    # verbose logging in this module
    export DEBUG_EXTRACT=1

    # verbose logging for just Redis operations in this module
    export DEBUG_REDIS_EXTRACT=1

    # verbose logging for just S3 operations in this module
    export DEBUG_S3_EXTRACT=1

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json

"""

import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.dataset_scrub_utils as scrub_utils
import analysis_engine.get_data_from_redis_key as redis_get
import analysis_engine.yahoo.consts as yahoo_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def extract_pricing_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_pricing_dataset

    Extract the Yahoo pricing data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    ds_id = work_dict.get('ticker')
    df_type = yahoo_consts.DATAFEED_PRICING_YAHOO
    df_str = yahoo_consts.get_datafeed_str_yahoo(df_type=df_type)
    redis_key = work_dict.get(
        'redis_key',
        work_dict.get('pricing', 'missing-redis-key'))
    s3_key = work_dict.get(
        's3_key',
        work_dict.get('pricing', 'missing-s3-key'))
    redis_host = work_dict.get(
        'redis_host',
        None)
    redis_port = work_dict.get(
        'redis_port',
        None)
    redis_db = work_dict.get(
        'redis_db',
        ae_consts.REDIS_DB)

    log.debug(
        f'{label} - {df_str} - start - redis_key={redis_key} s3_key={s3_key}')

    if not redis_host and not redis_port:
        redis_host = ae_consts.REDIS_ADDRESS.split(':')[0]
        redis_port = ae_consts.REDIS_ADDRESS.split(':')[1]

    df = None
    status = ae_consts.NOT_RUN
    try:
        redis_rec = redis_get.get_data_from_redis_key(
            label=label,
            host=redis_host,
            port=redis_port,
            db=redis_db,
            password=work_dict.get('password', None),
            key=redis_key,
            decompress_df=True)

        status = redis_rec['status']
        log.debug(
            f'{label} - {df_str} redis get data key={redis_key} '
            f'status={ae_consts.get_status(status=status)}')

        if status == ae_consts.SUCCESS:
            log.debug(f'{label} - {df_str} redis convert pricing to json')
            cached_dict = redis_rec['rec']['data']
            log.debug(f'{label} - {df_str} redis convert pricing to df')
            try:
                df = pd.DataFrame(
                    cached_dict,
                    index=[0])
            except Exception:
                log.debug(
                    f'{label} - {df_str} redis_key={redis_key} '
                    'no pricing df found')
                return ae_consts.EMPTY, None
            # end of try/ex to convert to df
            log.debug(
                f'{label} - {df_str} redis_key={redis_key} done '
                'convert pricing to df')
        else:
            log.debug(
                f'{label} - {df_str} did not find valid redis pricing '
                f'in redis_key={redis_key} '
                f'status={ae_consts.get_status(status=status)}')

    except Exception as e:
        log.debug(
            f'{label} - {df_str} - ds_id={ds_id} failed getting pricing from '
            f'redis={redis_host}:{redis_port}@{redis_db} '
            f'key={redis_key} ex={e}')
        return ae_consts.ERR, None
    # end of try/ex extract from redis

    log.debug(
        f'{label} - {df_str} ds_id={ds_id} extract scrub={scrub_mode}')

    scrubbed_df = scrub_utils.extract_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=df_type,
        msg_format='df={} date_str={}',
        ds_id=ds_id,
        df=df)

    status = ae_consts.SUCCESS

    return status, scrubbed_df
# end of extract_pricing_dataset


def extract_yahoo_news_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_yahoo_news_dataset

    Extract the Yahoo news data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    ds_id = work_dict.get('ticker')
    df_type = yahoo_consts.DATAFEED_NEWS_YAHOO
    df_str = yahoo_consts.get_datafeed_str_yahoo(df_type=df_type)
    redis_key = work_dict.get(
        'redis_key',
        work_dict.get('news', 'missing-redis-key'))
    s3_key = work_dict.get(
        's3_key',
        work_dict.get('news', 'missing-s3-key'))
    redis_host = work_dict.get(
        'redis_host',
        None)
    redis_port = work_dict.get(
        'redis_port',
        None)
    redis_db = work_dict.get(
        'redis_db',
        ae_consts.REDIS_DB)

    log.debug(
        f'{label} - {df_str} - start - redis_key={redis_key} s3_key={s3_key}')

    if not redis_host and not redis_port:
        redis_host = ae_consts.REDIS_ADDRESS.split(':')[0]
        redis_port = ae_consts.REDIS_ADDRESS.split(':')[1]

    df = None
    status = ae_consts.NOT_RUN
    try:
        redis_rec = redis_get.get_data_from_redis_key(
            label=label,
            host=redis_host,
            port=redis_port,
            db=redis_db,
            password=work_dict.get('password', None),
            key=redis_key,
            decompress_df=True)

        status = redis_rec['status']
        log.debug(
            f'{label} - {df_str} redis get data key={redis_key} '
            f'status={ae_consts.get_status(status=status)}')

        if status == ae_consts.SUCCESS:
            cached_dict = redis_rec['rec']['data']
            log.debug(f'{label} - {df_str} redis convert news to df')
            try:
                df = pd.DataFrame(
                    cached_dict)
            except Exception:
                log.debug(
                    f'{label} - {df_str} redis_key={redis_key} '
                    'no news df found')
                return ae_consts.EMPTY, None
            # end of try/ex to convert to df
            log.debug(
                f'{label} - {df_str} redis_key={redis_key} done '
                f'convert news to df')
        else:
            log.debug(
                f'{label} - {df_str} did not find valid redis news '
                f'in redis_key={redis_key} '
                f'status={ae_consts.get_status(status=status)}')

    except Exception as e:
        log.debug(
            f'{label} - {df_str} - ds_id={ds_id} failed getting news from '
            f'redis={redis_host}:{redis_port}@{redis_db} key={redis_key} '
            f'ex={e}')
        return ae_consts.ERR, None
    # end of try/ex extract from redis

    log.debug(f'{label} - {df_str} ds_id={ds_id} extract scrub={scrub_mode}')

    scrubbed_df = scrub_utils.extract_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=df_type,
        msg_format='df={} date_str={}',
        ds_id=ds_id,
        df=df)

    status = ae_consts.SUCCESS

    return status, scrubbed_df
# end of extract_yahoo_news_dataset


def extract_option_calls_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_option_calls_dataset

    Extract the Yahoo options calls for a ticker and
    return it as a ``pandas.Dataframe``

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = f'{work_dict.get("label", "extract")}-calls'
    ds_id = work_dict.get('ticker')
    df_type = yahoo_consts.DATAFEED_OPTIONS_YAHOO
    df_str = yahoo_consts.get_datafeed_str_yahoo(df_type=df_type)
    redis_key = work_dict.get(
        'redis_key',
        work_dict.get('calls', 'missing-redis-key'))
    s3_key = work_dict.get(
        's3_key',
        work_dict.get('calls', 'missing-s3-key'))
    redis_host = work_dict.get(
        'redis_host',
        None)
    redis_port = work_dict.get(
        'redis_port',
        None)
    redis_db = work_dict.get(
        'redis_db',
        ae_consts.REDIS_DB)

    log.debug(
        f'{label} - {df_str} - start - redis_key={redis_key} s3_key={s3_key}')

    if not redis_host and not redis_port:
        redis_host = ae_consts.REDIS_ADDRESS.split(':')[0]
        redis_port = ae_consts.REDIS_ADDRESS.split(':')[1]

    exp_date_str = None
    calls_df = None
    status = ae_consts.NOT_RUN
    try:
        redis_rec = redis_get.get_data_from_redis_key(
            label=label,
            host=redis_host,
            port=redis_port,
            db=redis_db,
            password=work_dict.get('password', None),
            key=redis_key,
            decompress_df=True)

        status = redis_rec['status']
        log.debug(
            f'{label} - {df_str} redis get data key={redis_key} '
            f'status={ae_consts.get_status(status=status)}')

        if status == ae_consts.SUCCESS:
            calls_json = None
            if 'calls' in redis_rec['rec']['data']:
                calls_json = redis_rec['rec']['data']['calls']
            else:
                calls_json = redis_rec['rec']['data']
            log.debug(f'{label} - {df_str} redis convert calls to df')
            exp_date_str = None
            try:
                calls_df = pd.read_json(
                    calls_json,
                    orient='records')
                exp_epoch_value = calls_df['expiration'].iloc[-1]
                exp_date_str = ae_utils.convert_epoch_to_datetime_string(
                    epoch=exp_epoch_value,
                    fmt=ae_consts.COMMON_DATE_FORMAT,
                    use_utc=True)
            except Exception:
                log.debug(
                    f'{label} - {df_str} redis_key={redis_key} '
                    'no calls df found')
                return ae_consts.EMPTY, None
            # end of try/ex to convert to df
            log.debug(
                f'{label} - {df_str} redis_key={redis_key} '
                f'calls={len(calls_df.index)} exp_date={exp_date_str}')
        else:
            log.debug(
                f'{label} - {df_str} did not find valid redis option calls '
                f'in redis_key={redis_key} '
                f'status={ae_consts.get_status(status=status)}')

    except Exception as e:
        log.debug(
            f'{label} - {df_str} - ds_id={ds_id} failed getting option calls '
            f'from redis={redis_host}:{redis_port}@{redis_db} '
            f'key={redis_key} ex={e}')
        return ae_consts.ERR, None
    # end of try/ex extract from redis

    log.debug(f'{label} - {df_str} ds_id={ds_id} extract scrub={scrub_mode}')

    scrubbed_df = scrub_utils.extract_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=df_type,
        msg_format='df={} date_str={}',
        ds_id=ds_id,
        df=calls_df)

    status = ae_consts.SUCCESS

    return status, scrubbed_df
# end of extract_option_calls_dataset


def extract_option_puts_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_option_puts_dataset

    Extract the Yahoo options puts for a ticker and
    return it as a ``pandas.Dataframe``

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = f'{work_dict.get("label", "extract")}-puts'
    ds_id = work_dict.get('ticker')
    df_type = yahoo_consts.DATAFEED_OPTIONS_YAHOO
    df_str = yahoo_consts.get_datafeed_str_yahoo(df_type=df_type)
    redis_key = work_dict.get(
        'redis_key',
        work_dict.get('puts', 'missing-redis-key'))
    s3_key = work_dict.get(
        's3_key',
        work_dict.get('puts', 'missing-s3-key'))
    redis_host = work_dict.get(
        'redis_host',
        None)
    redis_port = work_dict.get(
        'redis_port',
        None)
    redis_db = work_dict.get(
        'redis_db',
        ae_consts.REDIS_DB)

    log.debug(
        f'{label} - {df_str} - start - redis_key={redis_key} s3_key={s3_key}')

    if not redis_host and not redis_port:
        redis_host = ae_consts.REDIS_ADDRESS.split(':')[0]
        redis_port = ae_consts.REDIS_ADDRESS.split(':')[1]

    exp_date_str = None
    puts_df = None
    status = ae_consts.NOT_RUN
    try:
        redis_rec = redis_get.get_data_from_redis_key(
            label=label,
            host=redis_host,
            port=redis_port,
            db=redis_db,
            password=work_dict.get('password', None),
            key=redis_key,
            decompress_df=True)

        status = redis_rec['status']
        log.debug(
            f'{label} - {df_str} redis get data key={redis_key} '
            f'status={ae_consts.get_status(status=status)}')

        if status == ae_consts.SUCCESS:
            puts_json = None
            if 'puts' in redis_rec['rec']['data']:
                puts_json = redis_rec['rec']['data']['puts']
            else:
                puts_json = redis_rec['rec']['data']
            log.debug(f'{label} - {df_str} redis convert puts to df')
            try:
                puts_df = pd.read_json(
                    puts_json,
                    orient='records')
                exp_epoch_value = puts_df['expiration'].iloc[-1]
                exp_date_str = ae_utils.convert_epoch_to_datetime_string(
                    epoch=exp_epoch_value,
                    fmt=ae_consts.COMMON_DATE_FORMAT,
                    use_utc=True)
            except Exception:
                log.debug(
                    f'{label} - {df_str} redis_key={redis_key} '
                    'no puts df found')
                return ae_consts.EMPTY, None
            # end of try/ex to convert to df
            log.debug(
                f'{label} - {df_str} redis_key={redis_key} '
                f'puts={len(puts_df.index)} exp_date={exp_date_str}')
        else:
            log.debug(
                f'{label} - {df_str} did not find valid redis option puts '
                f'in redis_key={redis_key} '
                f'status={ae_consts.get_status(status=status)}')

    except Exception as e:
        log.debug(
            f'{label} - {df_str} - ds_id={ds_id} failed getting option puts '
            f'from redis={redis_host}:{redis_port}@{redis_db} '
            f'key={redis_key} ex={e}')
        return ae_consts.ERR, None
    # end of try/ex extract from redis

    log.debug(f'{label} - {df_str} ds_id={ds_id} extract scrub={scrub_mode}')

    scrubbed_df = scrub_utils.extract_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=df_type,
        msg_format='df={} date_str={}',
        ds_id=ds_id,
        df=puts_df)

    status = ae_consts.SUCCESS

    return status, scrubbed_df
# end of extract_option_puts_dataset



"""
Helper for extracting details from Celery task and
sending it to a slack webhook.

Supported environment variables:

::

    # slack webhook
    export SLACK_WEBHOOK=https://hooks.slack.com/services/

"""

import os
import json
import requests
import tabulate as tb
import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)
publish_msg_error = (
    'please add a SLACK_WEBHOOK environment variable to publish messages')
message_types = {
    'success': [{"color": "good", "title": "SUCCESS"}],
    'failure': [{"color": "danger", "title": "FAILED"}],
    'message': [{"title": "MESSAGE"}]
}


def post_success(msg,
                 jupyter=False,
                 block=False,
                 full_width=False):
    """Post a SUCCESS message to slack

    :param msg: A string, list, or dict to send to slack
    """
    return get_response('success',
                        msg,
                        jupyter=jupyter,
                        block=block,
                        full_width=full_width)


def post_failure(msg,
                 jupyter=False,
                 block=False,
                 full_width=False):
    """Post a FAILURE message to slack

    :param msg: A string, list, or dict to send to slack
    """
    return get_response('failure',
                        msg,
                        jupyter=jupyter,
                        block=block,
                        full_width=full_width)


def post_message(msg,
                 jupyter=False,
                 block=False,
                 full_width=False):
    """Post any message to slack

    :param msg: A string, list, or dict to send to slack
    """
    return get_response('message',
                        msg,
                        jupyter=jupyter,
                        block=block,
                        full_width=full_width)


def get_response(msg_type,
                 msg,
                 jupyter=False,
                 block=False,
                 full_width=False):
    """Attempt to create and send any message type to slack

    :param msg_type: A string of either 'success', 'failure', or 'message'
    :param msg: A string, list, or dict to send to slack
    """
    response = {'status': ae_consts.FAILED}
    if msg_type in message_types:
        if not os.getenv('SLACK_WEBHOOK', False):
            log.info(f'post_{msg_type} - {publish_msg_error}')
            return response
        if msg:
            attachments = [{"attachments": message_types[msg_type]}]
            fields = parse_msg(msg, block=block)
            if fields:
                if full_width:
                    attachments.append({"text": fields[0].pop("value")})
                else:
                    attachments[0]["attachments"][0]["fields"] = fields
                response = post(attachments, jupyter=jupyter)
    return response


def parse_msg(msg, block=False):
    """Create an array of fields for slack from the msg type

    :param msg: A string, list, or dict to massage for sending to slack
    """
    if type(msg) is str:
        if block:
            return [{"value": f"```{msg}```"}]
        return [{"value": msg}]
    elif type(msg) is list:
        if block:
            string_list = '\n'.join(f"{str(x)}" for x in msg)
            return [{"value": f"```{string_list}```"}]
        return [{"value": str(x)} for x in msg]
    elif type(msg) is dict:
        if block:
            string_dict = '\n'.join(
                f"{str(k)}: {str(v)}" for k, v in msg.items())
            return [{"value": f"```{string_dict}```"}]
        return [{"value": f"{str(k)}: {str(v)}"} for k, v in msg.items()]
    return None


def post(attachments, jupyter=False):
    """Send created attachments to slack

    :param attachments: Values to post to slack
    """
    SLACK_WEBHOOK = ae_consts.ev('SLACK_WEBHOOK', None)
    result = {'status': ae_consts.FAILED}
    if not os.getenv('SLACK_WEBHOOK', False):
        log.info(f'post - {publish_msg_error}')
        return result
    if attachments and SLACK_WEBHOOK:
        try:
            # if not jupyter:
            #     log.debug(f'Attempting to post attachments={attachments} '
            #               'to slack_webhook exists')
            for attachment in attachments:
                r = requests.post(SLACK_WEBHOOK, data=json.dumps(attachment))
                if str(r.status_code) == "200":
                    # log.info((
                    #   f'''Successful post of attachment={
                    #       attachment if not jupyter else
                    #       True if attachment else False} '''
                    #   'to slack_webhook'))
                    result['status'] = ae_consts.SUCCESS
                else:
                    log.error(
                        f'''Failed to post attachment={
                            attachment if not jupyter else
                            True if attachment else False} '''
                        f'with status_code={r.status_code}')
                    result['status'] = ae_consts.FAILED
                    break
        except Exception as e:
            log.error(
                f'''Failed to post attachments={
                    attachments if not jupyter else
                    True if attachments else False} '''
                f'with ex={e}')
            result['status'] = ae_consts.ERR
            result['err'] = e
    else:
        log.info(
            'Skipping post to slack due to missing '
            f'''attachments={
                attachments if not jupyter else
                True if attachments else False} or SLACK_WEBHOOK '''
            f'missing={False if SLACK_WEBHOOK else True}')
    return result


def post_df(
        df,
        columns=None,
        block=True,
        jupyter=True,
        full_width=True,
        tablefmt='github'):
    """post_df

    Post a ``pandas.DataFrame`` to Slack

    :param df: ``pandas.DataFrame`` object
    :param columns: ordered list of columns to for the table
                    header row
                    (``None`` by default)
    :param block: bool for
                  post as a Slack-formatted block ```like this```
                  (``True`` by default)
    :param jupyter: bool for
                    jupyter attachment handling
                    (``True`` by default)
    :param full_width: bool to ensure the width is preserved
                       the Slack message  (``True`` by default)
    :param tablefmt: string for table format (``github`` by default).
                     Additional format values can be found on:
                     https://bitbucket.org/astanin/python-tabulate
    """

    if not os.getenv('SLACK_WEBHOOK', False):
        log.info(f'post_df - {publish_msg_error}')
        return
    if not hasattr(df, 'index'):
        log.debug('post_df - no df ')
        return

    log.debug(
        f'post_df - df.index={len(df.index)} '
        f'columns={columns} fmt={tablefmt}')

    msg = None
    if columns:
        msg = tb.tabulate(
            df[columns],
            headers=columns,
            tablefmt=tablefmt)
    else:
        msg = tb.tabulate(
            df,
            tablefmt=tablefmt)
    # end of if/else

    post_success(
        msg=msg,
        block=block,
        jupyter=jupyter,
        full_width=full_width)
# end of post_df


def post_plot(plot,
              filename='tmp',
              title=None):
    """post_plot

    Post a matlibplot plot to Slack

    :param plot: matlibplot pyplot figure
    :param filename: filename of plot
    :param title: Title for slack plot postiing
    """

    result = {'status': ae_consts.FAILED}
    if not os.getenv('SLACK_ACCESS_TOKEN', False):
        log.info(
            'post_plot - please add a SLACK_ACCESS_TOKEN environment '
            'variable to publish plots')
        return result
    if not filename:
        log.info(
            'post_plot - no filename provided'
        )
        return result
    url = 'https://slack.com/api/files.upload'
    filename = f'{filename.split(".")[0]}.png'
    channels = [f'#{channel}' for channel in os.getenv(
        'SLACK_PUBLISH_PLOT_CHANNELS',
        'general').split(',')]
    try:
        log.info(f'post_plot - temporarily saving plot: {filename}')
        plot.savefig(filename)
        tmp_file = {
            'file': (filename, open(filename, 'rb'), 'png')
        }
        payload = {
            'channels':        channels,
            'filename':        filename,
            'title':           title if title else filename,
            'token':           os.getenv('SLACK_ACCESS_TOKEN'),
        }
        log.info(f'post_plot - sending payload: {payload} to url: {url}')
        r = requests.post(url,
                          params=payload,
                          files=tmp_file)
        log.info(f'post_plot - removing temporarily saved plot: {filename}')
        os.remove(filename)
        if str(r.status_code) == "200":
            log.info(f'post_plot - posted plot to slack channels: {channels}')
            result['status'] = ae_consts.SUCCESS
    except Exception as e:
        log.info(f'post_plot - failed with Exception: {e}')
        result['status'] = ae_consts.ERR
        result['err'] = e
    return result
# end of post_plot



"""
Show an algorithm dataset from file, s3 or redis

Supported Datasets:

- ``SA_DATASET_TYPE_ALGO_READY`` - Algorithm-ready datasets

**Supported environment variables**

::

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json
"""

import analysis_engine.consts as ae_consts
import analysis_engine.load_dataset as load_dataset
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def show_dataset(
        algo_dataset=None,
        dataset_type=ae_consts.SA_DATASET_TYPE_ALGO_READY,
        serialize_datasets=ae_consts.DEFAULT_SERIALIZED_DATASETS,
        path_to_file=None,
        compress=False,
        encoding='utf-8',
        redis_enabled=True,
        redis_key=None,
        redis_address=None,
        redis_db=None,
        redis_password=None,
        redis_expire=None,
        redis_serializer='json',
        redis_encoding='utf-8',
        s3_enabled=True,
        s3_key=None,
        s3_address=None,
        s3_bucket=None,
        s3_access_key=None,
        s3_secret_key=None,
        s3_region_name=None,
        s3_secure=False,
        slack_enabled=False,
        slack_code_block=False,
        slack_full_width=False,
        verbose=False):
    """show_dataset

    Show a supported dataset's internal structure and preview some
    of the values to debug mapping, serialization issues

    :param algo_dataset: optional - already loaded algorithm-ready dataset
    :param dataset_type: optional - dataset type
        (default is ``SA_DATASET_TYPE_ALGO_READY``)
    :param serialize_datasets: optional - list of dataset names to
        deserialize in the dataset
    :param path_to_file: optional - path to an algorithm-ready dataset
        in a file
    :param compress: optional - boolean flag for decompressing
        the contents of the ``path_to_file`` if necessary
        (default is ``False`` and algorithms
        use ``zlib`` for compression)
    :param encoding: optional - string for data encoding

    **(Optional) Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``True``)
    :param redis_key: string - key to save the data in redis
        (default is ``None``)
    :param redis_address: Redis connection string format: ``host:port``
        (default is ``localhost:6379``)
    :param redis_db: Redis db to use
        (default is ``0``)
    :param redis_password: optional - Redis password
        (default is ``None``)
    :param redis_expire: optional - Redis expire value
        (default is ``None``)
    :param redis_serializer: not used yet - support for future
        pickle objects in redis
    :param redis_encoding: format of the encoded key in redis

    **(Optional) Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``True``)
    :param s3_key: string - key to save the data in redis
        (default is ``None``)
    :param s3_address: Minio S3 connection string format: ``host:port``
        (default is ``localhost:9000``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``dev``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``trexaccesskey``)
    :param s3_secret_key: S3 Secret key
        (default is ``trex123321``)
    :param s3_region_name: S3 region name
        (default is ``us-east-1``)
    :param s3_secure: Transmit using tls encryption
        (default is ``False``)

    **(Optional) Slack arguments**

    :param slack_enabled: optional - boolean for
        publishing to slack
    :param slack_code_block: optional - boolean for
        publishing as a code black in slack
    :param slack_full_width: optional - boolean for
        publishing as a to slack using the full
        width allowed

    Additonal arguments

    :param verbose: optional - bool for increasing
        logging
    """

    use_ds = algo_dataset
    if not use_ds:
        log.info(
            f'loading from file={path_to_file} s3={s3_key} redis={redis_key}')
        use_ds = load_dataset.load_dataset(
            dataset_type=dataset_type,
            compress=compress,
            encoding=redis_encoding,
            path_to_file=path_to_file,
            s3_key=s3_key,
            s3_address=s3_address,
            s3_bucket=s3_bucket,
            s3_access_key=s3_access_key,
            s3_secret_key=s3_secret_key,
            s3_region_name=s3_region_name,
            s3_secure=s3_secure,
            redis_key=redis_key,
            redis_address=redis_address,
            redis_db=redis_db,
            redis_password=redis_password,
            redis_expire=redis_expire,
            redis_serializer=redis_serializer,
            serialize_datasets=serialize_datasets)

        if not use_ds:
            log.error(
                f'unable to load a dataset from file={path_to_file} '
                f's3={s3_key} redis={redis_key}')
            return None
    # load if not created

    if dataset_type == ae_consts.SA_DATASET_TYPE_ALGO_READY:
        print('-----------------------------------')
        for root_key in use_ds:
            print(root_key)
        all_dates = []
        all_ids = []
        first_node = None
        last_node = None
        end_nodes = []
        for root_key in use_ds:
            second_layer = use_ds[root_key]
            for ds in second_layer:
                if 'date' in ds:
                    if len(all_dates) == 0:
                        print('\ndates found in dataset')
                    cur_date = ds.get(
                        'date',
                        None)
                    if cur_date:
                        print(cur_date)
                        all_dates.append(cur_date)
                if not first_node:
                    first_node = ds
                end_nodes.append(ds)
                last_node = ds
                if 'id' in ds:
                    if len(all_ids) == 0:
                        print('\nids in the file')
                    cur_id = ds.get(
                        'id',
                        None)
                    if cur_id:
                        print(cur_id)
                        all_ids.append(cur_id)
        if first_node and last_node:
            show_first = {}
            for ds_key in first_node:
                if ds_key == 'data':
                    show_first[ds_key] = {}
                    for ds_name in first_node[ds_key]:
                        print(f'first_node has dataset with name: {ds_name}')
                        show_first[ds_key][ds_name] = 'EMPTY_DF'
                        if hasattr(
                                first_node[ds_key][ds_name],
                                'index'):
                            show_first[ds_key][ds_name] = (
                                'pd.DataFrame() rows='
                                f'{len(first_node[ds_key][ds_name].index)}')
                else:
                    show_first[ds_key] = first_node[ds_key]
            print(f'\nfirst node:\n{ae_consts.ppj(show_first)}\n')

            num_records = len(all_ids)
            cur_cell = num_records - 4
            for cur_node in end_nodes[-5:]:
                show_node = {}
                for ds_key in cur_node:
                    if ds_key == 'data':
                        show_node[ds_key] = {}
                        for ds_name in cur_node[ds_key]:
                            show_node[ds_key][ds_name] = 'EMPTY_DF'
                            if hasattr(
                                    cur_node[ds_key][ds_name],
                                    'index'):
                                show_node[ds_key][ds_name] = (
                                    'pd.DataFrame() rows='
                                    f'{len(cur_node[ds_key][ds_name].index)}')
                    else:
                        show_node[ds_key] = cur_node[ds_key]
                # end of show cur_node
                print(
                    f'node={cur_cell}/{num_records} values:'
                    f'\n{ae_consts.ppj(show_node)}\n')
                cur_cell += 1
            # end of end_nodes
        else:
            if not first_node:
                print('missing first node in dataset')
            if not last_node:
                print('missing last node in dataset')
        if len(all_dates) > 0:
            print(
                f'root_keys={use_ds} from {all_dates[0]} '
                f'to {all_dates[-1]}')
        else:
            print(f'root_keys={use_ds} missing dates')

        print('-----------------------------------')

    return use_ds
# end of show_dataset
